"""."""
import time as _time
from copy import deepcopy as _dcopy

import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from .device import Device as _Device, DeviceSet as _DeviceSet
from ..timesys.csdev import ETypes as _ETypes, Const as _TIConst, \
    get_hl_trigger_database as _get_hl_trigger_database
from ..search import HLTimeSearch as _HLTimeSearch
from ..util import get_bit as _get_bit


class EVG(_Device):
    """Device EVG."""

    DEVNAME = 'AS-RaMO:TI-EVG'
    StateMachine = _get_namedtuple('StateMachine', (
        'Initializing', 'Stopped', 'Continuous', 'Injection',
        'Preparing_Continuous', 'Preparing_Injection',
        'Restarting_Continuous'))

    PROPERTIES_DEFAULT = (
        'InjectionEvt-Sel', 'InjectionEvt-Sts', 'UpdateEvt-Cmd',
        'ContinuousEvt-Sel', 'ContinuousEvt-Sts', 'STATEMACHINE',
        'RepeatBucketList-SP', 'RepeatBucketList-RB',
        'BucketList-SP', 'BucketList-RB', 'BucketList-Mon',
        'BucketListLen-Mon', 'TotalInjCount-Mon', 'InjCount-Mon',
        'BucketListSyncStatus-Mon')

    def __init__(self, props2init='all'):
        """."""
        super().__init__(EVG.DEVNAME, props2init=props2init)

    @property
    def nrpulses(self):
        """Number of pulses to repeat Bucket List."""
        return self['RepeatBucketList-RB']

    @nrpulses.setter
    def nrpulses(self, value):
        """Set number of pulses to repeat Bucket List."""
        self['RepeatBucketList-SP'] = value

    @property
    def bucketlist_len(self):
        """Bucket List Length."""
        value = self['BucketListLen-Mon']
        return int(value) if value is not None else None

    @property
    def bucketlist_mon(self):
        """Implemented Bucket List."""
        return self['BucketList-Mon'][:self.bucketlist_len]

    @property
    def bucketlist(self):
        """Last setpoint accepted for Bucket List."""
        return self['BucketList-RB'][:self.bucketlist_len]

    @bucketlist.setter
    def bucketlist(self, value):
        """Set Bucket List."""
        self['BucketList-SP'] = _np.array(value, dtype=int)

    @property
    def bucketlist_sync(self):
        """Bucket list syncronized status."""
        return self['BucketListSyncStatus-Mon']

    @property
    def continuous_state(self):
        """."""
        return bool(self['ContinuousEvt-Sts'])

    @continuous_state.setter
    def continuous_state(self, value):
        """."""
        new_val = bool(value)
        if self.continuous_state != new_val:
            self['ContinuousEvt-Sel'] = int(new_val)

    @property
    def state_machine(self):
        """."""
        return self['STATEMACHINE']

    @property
    def state_machine_str(self):
        """."""
        return self.StateMachine._fields[self.state_machine]

    @property
    def injection_state(self):
        """."""
        return bool(self['InjectionEvt-Sts'])

    @injection_state.setter
    def injection_state(self, value):
        """."""
        new_val = bool(value)
        if self.injection_state != new_val:
            self['InjectionEvt-Sel'] = int(new_val)

    @property
    def injection_count_total(self):
        """."""
        return self['TotalInjCount-Mon']

    @property
    def injection_count(self):
        """."""
        return self['InjCount-Mon']

    def fill_bucketlist(self, stop, start=1, step=30, timeout=10):
        """Fill bucket list."""
        if (step > 0) and (start > stop):
            stop += 864
        if (step < 0) and (stop > start):
            stop -= 864
        value = _np.arange(start, stop, step)
        value = (value-1) % 864 + 1
        self.bucketlist = value
        rb_value = _np.zeros(864)
        rb_value[:len(value)] = value
        return self._wait('BucketList-RB', rb_value, timeout=timeout)

    def wait_injection_finish(self, timeout=10):
        """."""
        return self._wait(propty='InjectionEvt-Sts', value=0, timeout=timeout)

    def cmd_update_events(self, timeout=10):
        """."""
        val = self.continuous_state
        self['UpdateEvt-Cmd'] = 1
        _time.sleep(0.1)
        return self._wait(
            propty='ContinuousEvt-Sts', value=val, timeout=timeout)

    def cmd_turn_on_injection(self, timeout=10, wait_rb=False):
        """."""
        self.injection_state = 1
        pv2wait = 'InjectionEvt-' + ('Sts' if wait_rb else 'Sel')
        return self._wait(propty=pv2wait, value=1, timeout=timeout)

    def cmd_turn_off_injection(self, timeout=10, wait_rb=False):
        """."""
        self.injection_state = 0
        pv2wait = 'InjectionEvt-' + ('Sts' if wait_rb else 'Sel')
        return self._wait(propty=pv2wait, value=0, timeout=timeout)

    def cmd_turn_on_continuous(self, timeout=10, wait_rb=False):
        """."""
        self.continuous_state = 1
        pv2wait = 'ContinuousEvt-' + ('Sts' if wait_rb else 'Sel')
        return self._wait(propty=pv2wait, value=1, timeout=timeout)

    def cmd_turn_off_continuous(self, timeout=10, wait_rb=False):
        """."""
        self.continuous_state = 0
        pv2wait = 'ContinuousEvt-' + ('Sts' if wait_rb else 'Sel')
        return self._wait(propty=pv2wait, value=0, timeout=timeout)

    def set_nrpulses(self, value, timeout=10):
        """Set and wait number of pulses."""
        self['RepeatBucketList-SP'] = value
        return self._wait('RepeatBucketList-RB', value, timeout=timeout)


class Event(_Device):
    """Device Timing Event."""

    PROPERTIES_DEFAULT = (
        'Delay-SP', 'Delay-RB', 'DelayRaw-SP', 'DelayRaw-RB',
        'DelayType-Sel', 'DelayType-Sts', 'Mode-Sel', 'Mode-Sts',
        'Code-Mon', 'ExtTrig-Cmd',
        )

    MODES = _ETypes.EVT_MODES
    DELAYTYPES = ('Incr', 'Fixed')

    def __init__(self, evtname, props2init='all'):
        """."""
        super().__init__(
            EVG.DEVNAME + ':' + evtname, props2init=props2init)

    @property
    def mode(self):
        """."""
        return self['Mode-Sts']

    @mode.setter
    def mode(self, value):
        self._enum_setter('Mode-Sel', value, Event.MODES)

    @property
    def mode_str(self):
        """."""
        mode = self['Mode-Sts']
        if mode is not None:
            return Event.MODES[mode]
        return None

    @property
    def code(self):
        """."""
        return self['Code-Mon']

    @property
    def delay_type(self):
        """."""
        return self['DelayType-Sts']

    @delay_type.setter
    def delay_type(self, value):
        self._enum_setter('DelayType-Sel', value, Event.DELAYTYPES)

    @property
    def delay_type_str(self):
        """."""
        dlytyp = self['DelayType-Sts']
        if dlytyp is not None:
            return Event.DELAYTYPES[dlytyp]
        return None

    @property
    def delay(self):
        """."""
        return self['Delay-RB']

    @delay.setter
    def delay(self, value):
        self['Delay-SP'] = value

    @property
    def delay_raw(self):
        """."""
        return self['DelayRaw-RB']

    @delay_raw.setter
    def delay_raw(self, value):
        self['DelayRaw-SP'] = int(value)

    def cmd_external_trigger(self):
        """."""
        self['ExtTrig-Cmd'] = 1
        return True

    @property
    def is_in_inj_table(self):
        """."""
        return self.mode_str in Event.MODES[1:4]


class Trigger(_Device):
    """Device trigger."""

    STATES = ('Dsbl', 'Enbl')
    LOCKLL = ('Unlocked', 'Locked')
    POLARITIES = ('Normal', 'Inverse')

    PROPERTIES_DEFAULT = (
        'CtrldChannels-Cte', 'Delay-RB', 'Delay-SP', 'DelayRaw-RB',
        'DelayRaw-SP', 'DeltaDelay-RB', 'DeltaDelay-SP', 'DeltaDelayRaw-RB',
        'DeltaDelayRaw-SP', 'Duration-RB', 'Duration-SP', 'InInjTable-Mon',
        'LowLvlLock-Sel', 'LowLvlLock-Sts', 'LowLvlTriggers-Cte',
        'NrPulses-RB', 'NrPulses-SP', 'Polarity-Sel', 'Polarity-Sts',
        'Src-Sel', 'Src-Sts', 'State-Sel', 'State-Sts', 'Status-Mon',
        'StatusLabels-Cte', 'TotalDelay-Mon', 'TotalDelayRaw-Mon',
        'WidthRaw-RB', 'WidthRaw-SP')

    def __init__(self, trigname, props2init='all', auto_monitor_mon=False):
        """Init."""
        _database = _get_hl_trigger_database(trigname)
        all_props = tuple(_database)
        if props2init == 'all':
            props2init = all_props
        elif props2init is None:
            pass
        else:
            props2init = list(set(all_props) & set(props2init))
        self._source_options = _database['Src-Sel']['enums']
        super().__init__(
            trigname, props2init=props2init,
            auto_monitor_mon=auto_monitor_mon)

    @property
    def status(self):
        """Status."""
        return self['Status-Mon']

    @property
    def status_str(self):
        """Status string."""
        value = self.status
        strs = [d for i, d in enumerate(_TIConst.HLTrigStatusLabels)
                if _get_bit(value, i)]
        return ', '.join(strs) if strs else 'Ok'

    @property
    def status_labels(self):
        """Return Status labels of trigger.

        Returns:
            list: labels describing the possible statuses of the trigger.

        """
        pvo = self.pv_object('StatusLabels-Cte')
        return pvo.get(as_string=True).split('\n')

    @property
    def state(self):
        """State."""
        return self['State-Sts']

    @state.setter
    def state(self, value):
        self._enum_setter('State-Sel', value, Trigger.STATES)

    @property
    def state_str(self):
        """State string."""
        return Trigger.STATES[self['State-Sts']]

    @property
    def lock_low_level(self):
        """Lock low level status."""
        return self['LowLvlLock-Sts']

    @lock_low_level.setter
    def lock_low_level(self, value):
        self._enum_setter('LowLvlLock-Sel', value, Trigger.LOCKLL)

    @property
    def lock_low_level_str(self):
        """Lock low level status string."""
        return Trigger.LOCKLL[self['LowLvlLock-Sts']]

    @property
    def controlled_channels(self):
        """Return channels controlled by this trigger.

        Returns:
            list: names of the controlled channels.

        """
        pvo = self.pv_object('CtrldChannels-Cte')
        return pvo.get(as_string=True).split('\n')

    @property
    def low_level_triggers(self):
        """Return low level triggers controlled by this trigger.

        Returns:
            list: names of the controlled low level triggers.

        """
        pvo = self.pv_object('LowLvlTriggers-Cte')
        return pvo.get(as_string=True).split('\n')

    @property
    def source(self):
        """Source."""
        return self['Src-Sts']

    @source.setter
    def source(self, value):
        self._enum_setter('Src-Sel', value, self._source_options)

    @property
    def source_str(self):
        """Source string."""
        if self['Src-Sts'] is not None:
            return self._source_options[self['Src-Sts']]
        return

    @property
    def source_options(self):
        """Source options."""
        return self._source_options

    @property
    def duration(self):
        """Duration."""
        return self['Duration-RB']

    @duration.setter
    def duration(self, value):
        self['Duration-SP'] = value

    @property
    def width_raw(self):
        """Width of one pulse in hardware units."""
        return self['WidthRaw-RB']

    @width_raw.setter
    def width_raw(self, value):
        self['WidthRaw-SP'] = value

    @property
    def polarity(self):
        """Polarity."""
        return self['Polarity-Sts']

    @polarity.setter
    def polarity(self, value):
        self._enum_setter('Polarity-Sel', value, Trigger.POLARITIES)

    @property
    def polarity_str(self):
        """Polarity string."""
        return Trigger.POLARITIES[self['Polarity-Sts']]

    @property
    def nr_pulses(self):
        """Nr. of pulses."""
        return self['NrPulses-RB']

    @nr_pulses.setter
    def nr_pulses(self, value):
        self['NrPulses-SP'] = value

    @property
    def delay(self):
        """Delay."""
        return self['Delay-RB']

    @delay.setter
    def delay(self, value):
        self['Delay-SP'] = value

    @property
    def total_delay(self):
        """Total delay."""
        return self['TotalDelay-Mon']

    @property
    def delta_delay(self):
        """Return delta delay array.

        Returns:
            numpy.ndarray: delta delays.

        """
        return self['DeltaDelay-RB']

    @delta_delay.setter
    def delta_delay(self, value):
        if not isinstance(value, (_np.ndarray, list, tuple)):
            raise TypeError('Value must be a numpy.ndarray, list or tuple.')

        value = _np.array(value)
        size = self.delta_delay.size
        if value.size != size:
            raise TypeError(f'Size of value must be {size:d}.')
        self['DeltaDelay-SP'] = value

    @property
    def delay_raw(self):
        """Delay raw."""
        return self['DelayRaw-RB']

    @delay_raw.setter
    def delay_raw(self, value):
        self['DelayRaw-SP'] = value

    @property
    def total_delay_raw(self):
        """Total delay raw."""
        return self['TotalDelayRaw-Mon']

    @property
    def delta_delay_raw(self):
        """Return delta delay raw array.

        Returns:
            numpy.ndarray: delta delays raw.

        """
        return self['DeltaDelayRaw-RB']

    @delta_delay_raw.setter
    def delta_delay_raw(self, value):
        if not isinstance(value, (_np.ndarray, list, tuple)):
            raise TypeError('Value must be a numpy.ndarray, list or tuple.')

        value = _np.array(value)
        size = self.delta_delay_raw.size
        if value.size != size:
            raise TypeError(f'Size of value must be {size:d}.')
        self['DeltaDelayRaw-SP'] = value

    @property
    def is_in_inj_table(self):
        """Is in Injection table."""
        return self['InInjTable-Mon']

    def cmd_enable(self, timeout=3):
        """Command enable."""
        self.state = 1
        return self._wait('State-Sts', 1, timeout)

    def cmd_disable(self, timeout=3):
        """Command disable."""
        self.state = 0
        return self._wait('State-Sts', 0, timeout)

    def cmd_lock_low_level(self, timeout=3):
        """Lock low level IOCs state."""
        self.lock_low_level = 1
        return self._wait('LowLvlLock-Sts', 1, timeout)

    def cmd_unlock_low_level(self, timeout=3):
        """Unlock low level IOCs state."""
        self.lock_low_level = 0
        return self._wait('LowLvlLock-Sts', 0, timeout)


class HLTiming(_DeviceSet):
    """."""

    SEARCH = _HLTimeSearch

    def __init__(
            self, controlled_trigs='all', trigs_props2init='all',
            evts_props2init='all'):
        """."""
        self.evg = EVG()
        evs = self.SEARCH.get_configurable_hl_events()
        self.events = {
            ev: Event(ev, props2init=evts_props2init) for ev in evs.keys()}

        self._triggernames_all = self.SEARCH.get_hl_triggers()
        trigs = self._triggernames_all
        if isinstance(controlled_trigs, (list, tuple)):
            trigs = sorted(set(controlled_trigs) & set(self.triggernames_all))
        self.triggers = {
            t: Trigger(t, props2init=trigs_props2init) for t in trigs}
        self._trigs_props2init = trigs_props2init

        devs = [self.evg, ]
        devs += list(self.events.values())
        devs += list(self.triggers.values())
        super().__init__(devs, devname='AS-Glob:TI-HLTiming')

    @property
    def triggernames_controlled(self):
        """Names of the triggers controlled by this class."""
        return list(self.triggers)

    @property
    def triggernames_all(self):
        """Names of all the possible high level triggers."""
        return _dcopy(self._triggernames_all)

    @property
    def is_full(self):
        """Return True if this class controls all triggers."""
        return not bool(set(self._triggernames_all) - set(self.triggers))

    def add_trigger(self, trigname):
        """Add trigger to the list of controlled triggers.

        Args:
            trigname (str): Trigger name to add

        Returns:
            bool: whether addition was sucessful.

        """
        if trigname not in set(self._triggernames_all):
            return False
        if trigname not in self.triggers:
            self.triggers[trigname] = Trigger(
                trigname, props2init=self._trigs_props2init)
        return True

    def get_mapping_events2triggers(self) -> dict:
        """."""
        map_ = {tn: tr.source_str for tn, tr in self.triggers.items()}
        inv_map = {ev: list() for ev in set(map_.values())}
        for tn, ev in map_.items():
            inv_map[ev].append(tn)
        return inv_map

    def get_mapping_injtable2events(self) -> dict:
        """."""
        map_evt2table = {n: o.mode_str for n, o in self.events.items()}
        map_table2evt = {}
        for k, v in map_evt2table.items():
            map_table2evt[v] = map_table2evt.get(v, []) + [k]
        return map_table2evt

    def change_triggers_source(
            self, trigs, new_src='Linac', printlog=True) -> list:
        """."""
        notchanged = list()
        for tn in trigs:
            tr = self.triggers.get(tn)
            if tr is None:
                notchanged.append(tn)
                if not printlog:
                    continue
                print(f'{tn:25s} -> No Change: {tn:s} is not controlled.')
                continue

            if new_src not in tr.source_options:
                notchanged.append(tn)
                if not printlog:
                    continue
                print(f'{tn:25s} -> No Change: {new_src:s} is not an option.')
                continue

            dly_newsrc = 0
            if new_src in self.events:
                dly_newsrc = self.events[new_src].delay_raw

            dly_oldsrc = 0
            old_src = tr.source_str
            if old_src in self.events:
                dly_oldsrc = self.events[old_src].delay_raw

            dly = tr.delay_raw
            delta_dly = dly_oldsrc
            delta_dly -= dly_newsrc
            dly += delta_dly
            if dly < 0:
                notchanged.append(tn)
                if printlog:
                    print(f'{tn:25s} -> No Change: total delay not constant!')
                continue

            tr.delay_raw = dly
            tr.source = new_src
            if printlog:
                print(f'{tn:25s} -> Change OK: .')
        return notchanged

    def change_event_delay(self, new_dly, event='Linac', printlog=True):
        """."""
        if not self.is_full:
            if printlog:
                print('Aborted: class does not control all triggers.')
            return False

        if event not in self.events:
            if printlog:
                print(f'{event} is not a valid event!')
            return False
        new_dly = int(new_dly)
        old_dly = self.events[event].delay_raw
        dlt_dly = old_dly - new_dly

        trigs = self.get_mapping_events2triggers()[event]
        for trn in trigs:
            dly = self.triggers[trn].delay_raw + dlt_dly
            if dly < 0:
                if printlog:
                    print(f'cannot change delay: {trn:s} would change!')
                return False

        for trn in trigs:
            self.triggers[trn].delay_raw += dlt_dly
        self.events[event].delay_raw = new_dly
        if printlog:
            print('Delay changed!')
        return True

    def print_injtable_mapping(self, only_enabled=False):
        """."""
        map_evt2trig = self.get_mapping_events2triggers()
        map_table2evt = self.get_mapping_injtable2events()
        tabs = {'Continuous', 'Injection', 'OneShot'}
        tabs &= map_table2evt.keys()
        tabs = sorted(tabs)

        dlys = []
        for tab in tabs:
            for evt in map_table2evt[tab]:
                for name in map_evt2trig.get(evt, []):
                    obj = self.triggers[name]
                    if only_enabled and not obj.enabled:
                        continue
                    dlys.append([obj.total_delay, name, evt, tab])
        dlys = sorted(dlys)

        tmpl = ' {:^30s} |' * len(tabs)
        print(('{:^12s} |' + tmpl).format('Delay [ms]', *tabs))
        print('-'*(12+33*len(tabs) + 2))
        for dly, trg, evt, tab in dlys:
            stgs = [''] * len(tabs)
            stgs[tabs.index(tab)] = trg
            print(('{:>12.6f} |' + tmpl).format(dly/1e3, *stgs))



"""Lattice survey module."""

import pyaccel
import pymodels
from ..namesys import join_name as _join_name, SiriusPVName as _PVName
from ..clientweb import crates_mapping as _crates_mapping


def run_all():
    """."""
    generate_bpm_static_table()
    generate_model_static_table()


def generate_model_static_table():
    """."""
    disclaimer = \
        "# This file was generated automatically from the data of the \n"\
        "# models of the accelerators in the pymodels repository by the \n"\
        "# function: \n"\
        "#    siriuspy.optics.lattice_survey.generate_model_static_table.\n"\
        "#\n"\
        "# If any model change, please, run the script\n"\
        "# again and copy the generated file to replace this one.\n"

    filename = 'magnets-model-data.txt'

    with open(filename, 'w') as f:
        f.write(disclaimer)
        f.write('\n\n#Magnets data \n')
        f.write('#-----------')

    model = pymodels.si
    acc = model.create_accelerator()
    _append_mag_data(filename, model, acc, 'Storage Ring (SI)', 'SI')

    model = pymodels.bo
    acc = model.create_accelerator()
    ind = pyaccel.lattice.find_indices(acc, 'fam_name', 'InjSept')
    acc = pyaccel.lattice.shift(acc, ind[0])
    _append_mag_data(filename, model, acc, 'Booster (BO)', 'BO')

    model = pymodels.tb
    acc, *_ = model.create_accelerator(add_from_li_triplets=False)
    _append_mag_data(
        filename, model, acc, 'Booster Transport Line (TB)', 'TB')

    model = pymodels.ts
    acc, *_ = model.create_accelerator()
    _append_mag_data(filename, model, acc, 'SR Transport Line (TS)', 'TS')


def _conv_dev_inst(section, sub, mag_tp, inst):
    inst_conv = {
        # additional correctors were added in TS
        'TS-01': {
            'CV': {'2': ('CV', '1E2'), '3': ('CV', '2')},
            },
        'TS-02': {
            'CV': {'1': ('CV', '0'), '2': ('CV', '')},
            },
        'TS-04': {
            'CV': {
                '1': ('CV', '0'), '2': ('CV', '1'),
                '3': ('CV', '1E2'), '4': ('CV', '2')},
            },
        # TB skew quad in the model converted to ch
        'TB-04': {
            'CH': {'1': ('CH', '')},
            'CH': {'2': ('QS', '')},
            },
        'TB-04': {
            'CH': {'': ('CH', '1')},
            'QS': {'': ('CH', '2')},
            },
        }
    secsub = section + '-' + sub
    inst_conv_ = inst_conv
    if secsub in inst_conv_:
        inst_conv_ = inst_conv_[secsub]
        if mag_tp in inst_conv_:
            inst_conv_ = inst_conv_[mag_tp]
            if inst in inst_conv_:
                return inst_conv_[inst]
    return mag_tp, inst


def _conv_id_correctors(mag_tp):
    if mag_tp in ('IDCH', 'IDCV', 'IDQS'):
        # NOTE: this would issue wrong instances if there were
        # other CH or CV magnets in insertion device straights, which
        # is not the case.
        mag_tp = mag_tp[2:]
    return mag_tp


def _conv_pulsed_magnets(mag_tp, dis):
    if 'Inj' in mag_tp or 'Eje' in mag_tp or 'Ping' in mag_tp:
        dis = 'PM'
    return dis


def _append_mag_data(filename, model, acc, label, section):
    fam_data = model.get_family_data(acc)
    pos = pyaccel.lattice.find_spos(acc)

    args = (acc, ) if model == pymodels.tb else ()
    mag_tps = []
    mag_tps.extend(model.families.families_dipoles(*args))
    mag_tps.extend(model.families.families_pulsed_magnets(*args))
    mag_tps.extend(model.families.families_quadrupoles(*args))
    mag_tps.extend(model.families.families_horizontal_correctors(*args))
    mag_tps.extend(model.families.families_vertical_correctors(*args))
    mag_tps.extend(model.families.families_sextupoles(*args))
    mag_tps.extend(model.families.families_skew_correctors(*args))

    mag_data = dict()
    for mag_tp in mag_tps:
        inds = fam_data[mag_tp]['index']
        subs = fam_data[mag_tp]['subsection']
        insts = fam_data[mag_tp]['instance']
        for ind, inst, sub in zip(inds, insts, subs):
            val = (pos[ind[-1]+1] + pos[ind[0]]) / 2  # elem position
            # special conversions
            mag_tp_, inst_, dis_ = mag_tp, inst, 'MA'
            mag_tp_ = _conv_id_correctors(mag_tp_)
            dis_ = _conv_pulsed_magnets(mag_tp_, dis_)
            mag_tp_, inst_ = _conv_dev_inst(section, sub, mag_tp_, inst_)
            name = _join_name(
                sec=section, dis=dis_, dev=mag_tp_, sub=sub, idx=inst_)
            mag_data[name] = val

    # order by spos
    mags = list(mag_data.keys())
    mpos = list(mag_data.values())
    mpos, mags = zip(*sorted(zip(mpos, mags)))

    with open(filename, 'a') as f:
        f.write('\n\n\n# '+label+'\n')
        f.write('#'+57*'-' + '\n')
        f.write("#{mag:20s} {pos:^20s}\n".format(
            mag='Name', pos='Position @ center [m]'))
        f.write('#'+57*'-' + '\n')
        for mag, pos in zip(mags, mpos):
            f.write(f'{mag:20s} {pos:>20.4f}\n')


def generate_bpm_static_table():
    """."""
    disclaimer = \
        "# This file was generated automatically from the data of the files\n"\
        "# in the subfolder Mapeamento_placas_MicroTCA_vs_BPMs and the\n"\
        "# models of the accelerators in the pymodels repository by the\n"\
        "# function: \n"\
        "#      siriuspy.optics.lattice_survey.generate_bpm_static_table.\n"\
        "#\n"\
        "# If the data in the mentioned subfolder change, please, run the\n"\
        "# script again and copy the generated file to replace this one.\n"

    filename = 'bpms-data.txt'
    with open(filename, 'w') as f:
        f.write(disclaimer)
        f.write('\n\n#BPMs data \n')
        f.write('#-----------')

    all_bpms = _get_all_bpms()

    model = pymodels.si
    acc = model.create_accelerator()
    _append_bpm_data_bl(filename, acc, all_bpms, 'Beam Lines (BL)')
    label = 'Storage Ring (SR)'
    _append_bpm_data(filename, model, acc, label, 'SI', fam='BPM')
    label = 'Storage Ring IDs (SR)'
    _append_bpm_data(filename, model, acc, label, 'SI', fam='IDBPM')

    model = pymodels.bo
    acc = model.create_accelerator()
    _append_bpm_data(filename, model, acc, 'Booster (BO)', 'BO')

    model = pymodels.tb
    acc, *_ = model.create_accelerator()
    idx = pyaccel.lattice.find_indices(acc, 'fam_name', 'Spect')[0]
    acc = acc[idx+1:]
    _append_bpm_data(filename, model, acc, 'BO Transport Line', 'TB')

    model = pymodels.ts
    acc, *_ = model.create_accelerator()
    _append_bpm_data(filename, model, acc, 'SR Transport Line', 'TS')


def _append_bpm_data(filename, model, acc, label, section, fam='BPM'):
    fam_data = model.get_family_data(acc)
    pos = pyaccel.lattice.find_spos(acc)

    bpm_data = fam_data[fam]
    inds = [i[0] for i in bpm_data['index']]
    subs = bpm_data['subsection']
    insts = bpm_data['instance']
    bpms, bpos = [], []
    for ind, inst, sub in zip(inds, insts, subs):
        name = _join_name(sec=section, dis='DI', dev='BPM', sub=sub, idx=inst)
        bpms.append(name)
        bpos.append(pos[ind])
    _write_to_file(filename, bpms, bpos, label)


def _append_bpm_data_bl(filename, acc, all_bpms, label):
    pos = pyaccel.lattice.find_spos(acc)

    mcs = pyaccel.lattice.find_indices(acc, 'fam_name', 'mc')

    bpms_bc = [bpm for bpm in all_bpms if bpm.sub.endswith('BCFE')]
    secs = [int(bpm.sub[:2]) for bpm in bpms_bc]
    bpos_bc = [pos[mcs[ss-1]] for ss in secs]

    mia = pyaccel.lattice.find_indices(acc, 'fam_name', 'mia')
    mib = pyaccel.lattice.find_indices(acc, 'fam_name', 'mib')
    mip = pyaccel.lattice.find_indices(acc, 'fam_name', 'mip')
    mis = sorted(mia + mib + mip)

    end = ('SAFE', 'SBFE', 'SPFE')
    bpms_mi = sorted([bpm for bpm in all_bpms if bpm.sub.endswith(end)])
    secs = [int(bpm.sub[:2]) for bpm in bpms_mi]
    bpos_mi = [pos[mis[ss-1]] for ss in secs]

    bpms = bpms_mi + bpms_bc
    bpos = bpos_mi + bpos_bc
    data = sorted(zip(bpos, bpms))
    bpos, bpms = list(zip(*data))
    _write_to_file(filename, bpms, bpos, label)


def _get_all_bpms():
    data = _crates_mapping()
    bpms = set()
    for line in data.splitlines():
        line = line.strip()
        if not line or line[0] == '#':
            continue  # empty line
        dev, *_ = line.split()
        dev = _PVName(dev)
        if dev.dev in ('BPM', 'PBPM'):
            bpms.add(dev)
    return bpms


def _write_to_file(fname, bpms, pos, label):
    with open(fname, 'a') as fil:
        fil.write('\n\n\n# '+label+'\n')
        fil.write('#'+57*'-' + '\n')
        fil.write("#{bpm:20s} {pos:^15s}\n".format(
            bpm='Name', pos='Position [m]'))
        fil.write('#'+57*'-' + '\n')
        for bpm, p in zip(bpms, pos):
            fil.write(f"{bpm:20s} {p:^15.4f}\n")



"""Module with main IOC Class."""

import time as _time
import logging as _log
import numpy as _np

from pcaspy import Alarm as _Alarm
from pcaspy import Severity as _Severity

from .. import util as _util

from .meas import MeasDVF


class App:
    """BL Image Processing IOC Application."""

    CHECK_IMG_ACQUIRE = True

    _MON_PVS_2_IMGFIT = {
        # These PVs are updated at evry image processing
        # --- image intensity ---
        'ImgSizeX-Mon': ('fitx', 'size'),
        'ImgSizeY-Mon': ('fity', 'size'),
        'ImgIntensityMin-Mon': 'intensity_min',
        'ImgIntensityMax-Mon': 'intensity_max',
        'ImgIntensitySum-Mon': 'intensity_sum',
        'ImgIsSaturated-Mon': 'is_saturated',
        # --- image projection ---
        'ImgProjX-Mon': ('fitx', 'data'),
        'ImgProjY-Mon': ('fity', 'data'),
        'ImgIsWithBeam-Mon': 'is_with_image',
        # --- roix ---
        'ImgROIX-RB': ('fitx', 'roi'),
        'ImgROIXCenter-Mon': ('fitx', 'roi_center'),
        'ImgROIXFWHM-Mon': ('fitx', 'roi_fwhm'),
        'ImgROIXProj-Mon': ('fitx', 'roi_proj'),
        # --- roix_fit ---
        'ImgROIXFitAmplitude-Mon': ('fitx', 'roi_amplitude'),
        'ImgROIXFitMean-Mon': ('fitx', 'roi_mean'),
        'ImgROIXFitSigma-Mon': ('fitx', 'roi_sigma'),
        'ImgROIXFitError-Mon': ('fitx', 'roi_fit_error'),
        # --- roixy ---
        'ImgROIY-RB': ('fity', 'roi'),
        'ImgROIYCenter-Mon': ('fity', 'roi_center'),
        'ImgROIYFWHM-Mon': ('fity', 'roi_fwhm'),
        'ImgROIYProj-Mon': ('fity', 'roi_proj'),
        # --- roiy_fit ---
        'ImgROIYFitAmplitude-Mon': ('fity', 'roi_amplitude'),
        'ImgROIYFitMean-Mon': ('fity', 'roi_mean'),
        'ImgROIYFitSigma-Mon': ('fity', 'roi_sigma'),
        'ImgROIYFitError-Mon': ('fity', 'roi_fit_error'),
        # --- gauss2d fit ---
        'ImgFitAngle-Mon': 'angle',
        'ImgFitSigma1-Mon': 'sigma1',
        'ImgFitSigma2-Mon': 'sigma2',
    }

    _INIT_PVS_2_IMGFIT = {
        # These are either constant PVs or readback PVs whose
        # initializations need external input
        'ImgROIX-RB': ('fitx', 'roi'),
        'ImgROIY-RB': ('fity', 'roi'),
        'ImgROIX-SP': ('fitx', 'roi'),
        'ImgROIY-SP': ('fity', 'roi'),
    }

    def __init__(self, driver=None, const=None):
        """Initialize the instance."""
        self._driver = driver
        self._const = const
        self._database = const.get_database()
        self._heartbeat = 0
        self._timestamp_last_update = _time.time()
        self._init_driver_flag = False

        # create measurement object
        self._meas = self._create_meas()

        # print info about the IOC
        dbase = self._database
        _util.print_ioc_banner(
            ioc_name='BL ImgProc IOC',
            db=dbase,
            description='Image Processing IOC (FAC)',
            version=dbase['ImgVersion-Cte']['value'],
            prefix=const.devname)

        # add epics app callback to measurement
        self._meas.callback = self.update_driver

    @property
    def const(self):
        """."""
        return self._const

    @property
    def driver(self):
        """."""
        return self._driver

    @driver.setter
    def driver(self, value):
        """."""
        self._driver = value

    @property
    def meas(self):
        """."""
        return self._meas

    @property
    def heartbeat(self):
        """."""
        return self._heartbeat

    def increment_heartbeat(self):
        """."""
        self._heartbeat += 1

    def process(self, interval):
        """Run continuously in the main thread."""
        # update heartbeat
        self.increment_heartbeat()
        self._write_pv('ImgTimestampUpdate-Mon', _time.time())

        # if enabled, check if image acquisition is not working and set it
        if App.CHECK_IMG_ACQUIRE:
            self._check_acquisition_timeout()

        _time.sleep(interval)

    def write(self, reason, value):
        """Write value in objects and database."""
        pv_is_writable = reason.endswith('-SP') \
            or reason.endswith('-Sel') \
            or reason.endswith('-Cmd')
        if not pv_is_writable:
            self._log_warning(f'PV {reason} is not writable!')
            return False

        res = self._write_roi(reason, value)
        if res is not None:
            return res

        res = self._write_fwhm_factor(reason, value)
        if res is not None:
            return res

        res = self._write_iswithbeam_threshold(reason, value)
        if res is not None:
            return res

        res = self._write_update_roi_with_fwhm(reason, value)
        if res is not None:
            return res

        res = self._write_use_svd4theta(reason, value)
        if res is not None:
            return res

        res = self._write_dvf_reset(reason, value)
        if res is not None:
            return res

        res = self._write_dvf_acquire(reason, value)
        if res is not None:
            return res

        return True

    def init_driver(self):
        """Initialize PVs at startup."""
        self._init_driver_flag = True
        self._write_pv('ImgDVFSizeX-Cte', self.meas.dvf_sizex)
        self._write_pv('ImgDVFSizeY-Cte', self.meas.dvf_sizey)

        msgfmt_nok = 'PV {} could not be initialized!'
        msgfmt_ok = 'PV {} initialized.'
        for pvname, attr in App._INIT_PVS_2_IMGFIT.items():
            if self.meas.status == self.meas.STATUS_SUCCESS:
                # update epics db successfully
                value = self._conv_imgattr2value(attr)
                if value is None:
                    self._log_warning(msgfmt_nok.format(pvname))
                    self._write_pv_failed(pvname)
                    self._init_driver_flag = False
                else:
                    _log.info(msgfmt_ok.format(pvname))
                    self._write_pv(pvname, value)
            else:
                self._log_warning(msgfmt_nok.format(pvname))
                self._write_pv_failed(pvname)
                self._init_driver_flag = False

    def update_driver(self):
        """Update all parameters at every image PV callback."""
        self._timestamp_last_update = _time.time()

        if not self._init_driver_flag:
            self.init_driver()

        if not self.meas.image2dfit.is_with_image:
            self._write_pv('ImgIsWithBeam-Mon', 0)

        invalid_fitx, invalid_fity = [False]*2
        with_beam = self.meas.image2dfit.is_with_image

        for pvname, attr in App._MON_PVS_2_IMGFIT.items():
            # check if is roi_rb and if it needs updating
            if pvname in ('ImgROIX-RB', 'ImgROIY-RB'):
                if not self.meas.update_roi_with_fwhm:
                    continue

            # if no beam, return if PV is of fit type
            if not with_beam and 'Fit' in pvname and 'ProcTime' not in pvname:
                return

            # get image attribute value
            value = self._conv_imgattr2value(attr)
            if value is None:
                msg = f'PV {pvname} could not be updated with None value!'
                self._log_warning(msg)
                continue

            # check if fit is valid and update value
            invalid_fit, new_value = self._check_invalid_fit(pvname, value)
            invalid_fitx |= invalid_fit if 'XFit' in pvname else False
            invalid_fity |= invalid_fit if 'YFit' in pvname else False

            # update epics db
            if self.meas.status == self.meas.STATUS_SUCCESS:
                self._write_pv(pvname, new_value)
            else:
                self._write_pv_failed(pvname)

        if invalid_fitx:
            msgerr = 'Invalid ROIXFit'
            self._log_warning(msgerr)
            self._write_pv_log(msgerr)
        if invalid_fity:
            msgerr = 'Invalid ROIYFit'
            self._log_warning(msgerr)
            self._write_pv_log(msgerr)

        if self.meas.status != self.meas.STATUS_SUCCESS:
            self._log_warning(self.meas.status)
            self._write_pv_log(self.meas.status)
        else:
            if self.meas.proc_time is not None:
                self._write_pv('ImgFitProcTime-Mon', self.meas.proc_time)
            self._write_pv('ImgDVFStatus-Mon', self.meas.status_dvf)

    def _check_acquisition_timeout(self):
        """."""
        interval = _time.time() - self._timestamp_last_update
        if self.meas.acquisition_timeout(interval):
            msgfmt_nok = 'DVF Image update timeout!'
            self._log_warning(msgfmt_nok)
            self._write_pv_log(msgfmt_nok)
            self._timestamp_last_update = _time.time()

    def _check_invalid_fit(self, pvname, value):
        """."""
        if 'XFit' in pvname:
            invalid_fit = self.meas.image2dfit.fitx.invalid_fit
        elif 'YFit' in pvname:
            invalid_fit = self.meas.image2dfit.fity.invalid_fit
        elif 'FitAngle' in pvname:
            invalid_fit = \
                self.meas.image2dfit.fitx.invalid_fit or \
                self.meas.image2dfit.fity.invalid_fit
        else:
            invalid_fit = False

        if invalid_fit:
            new_value = 0
        else:
            new_value = value
        return invalid_fit, new_value

    def _create_meas(self):
        # build arguments
        fwhmx_factor = \
            float(self._database['ImgROIXUpdateWithFWHMFactor-RB']['value'])
        fwhmy_factor = \
            float(self._database['ImgROIYUpdateWithFWHMFactor-RB']['value'])
        roi_with_fwhm = \
            float(self._database['ImgROIUpdateWithFWHM-Sts']['value'])
        intensity_threshold = \
            int(self._database['ImgIsWithBeamThreshold-RB']['value'])
        use_svd4theta = \
            int(self._database['ImgFitAngleUseCMomSVD-Sts']['value'])

        # create object
        meas = MeasDVF(
            self.const,
            fwhmx_factor=fwhmx_factor, fwhmy_factor=fwhmy_factor,
            roi_with_fwhm=roi_with_fwhm,
            intensity_threshold=intensity_threshold,
            use_svd4theta=use_svd4theta,
        )
        return meas

    def _write_pv(self, pvname, value=None, success=True):
        """."""
        if success:
            if isinstance(value, (bool, _np.bool, _np.bool_)):
                value = 1 if value else 0
            try:
                self._driver.setParam(pvname, value)
                self._driver.updatePV(pvname)
            except TypeError:
                _log.warning(
                    '_write_pv: error in updatePV for ', pvname, value)
            self._driver.setParamStatus(
                pvname, _Alarm.NO_ALARM, _Severity.NO_ALARM)
        else:
            _log.debug('{}: updated with alarm'.format(pvname))
            self._driver.setParamStatus(
                pvname, _Alarm.TIMEOUT_ALARM, _Severity.INVALID_ALARM)

    def _write_pv_failed(self, pvname):
        # update epics db failure
        self._write_pv(pvname, success=False)
        # update Log
        self._write_pv_log(self.meas.status)

    def _write_pv_log(self, message, success=True):
        """."""
        message = f'[{self.heartbeat}] ' + message
        self._write_pv('ImgLog-Mon', message, success)

    def _write_pv_sp_rb(self, reason, value):
        # update SP/Sel
        self._write_pv(reason, value)

        # update RB/Sts
        reason = reason.replace('-SP', '-RB').replace('-Sel', '-Sts')
        self._write_pv(reason, value)

    def _write_roi(self, reason, value):
        if reason not in ('ImgROIX-SP', 'ImgROIY-SP'):
            return None
        if 'X' in reason:
            self.meas.set_roix(value)
        else:
            self._meas.set_roiy(value)

        if self.meas.status == self.meas.STATUS_SUCCESS:
            self._write_pv_sp_rb(reason, value)
            return True
        else:
            msg = '{}: could not write value {}'.format(reason, value)
            self._log_warning(msg)
            self._driver.setParam('ImgLog-Mon', msg)
            self._driver.updatePV('ImgLog-Mon')
            return False

    def _write_use_svd4theta(self, reason, value):
        if reason != 'ImgFitAngleUseCMomSVD-Sel':
            return None

        self.meas.set_use_svd4theta(value)

        if self.meas.status == self.meas.STATUS_SUCCESS:
            self._write_pv_sp_rb(reason, value)
            return True
        else:
            msg = '{}: could not write value {}'.format(reason, value)
            self._log_warning(msg)
            self._driver.setParam('ImgLog-Mon', msg)
            self._driver.updatePV('ImgLog-Mon')
            return False

    def _write_iswithbeam_threshold(self, reason, value):
        if reason != 'ImgIsWithBeamThreshold-SP':
            return None

        # set threshold
        self.meas.set_intensity_threshold(value)

        if self.meas.status == self.meas.STATUS_SUCCESS:
            self._write_pv_sp_rb(reason, value)
            return True
        else:
            msg = '{}: could not write value {}'.format(reason, value)
            self._log_warning(msg)
            self._driver.setParam('ImgLog-Mon', msg)
            self._driver.updatePV('ImgLog-Mon')
            return False

    def _write_dvf_reset(self, reason, value):
        """."""
        if reason != 'ImgDVFReset-Cmd':
            return None

        # set acquire
        if self.meas.reset_dvf():
            value = self._driver.getParam(reason)
            self._driver.setParam(reason, value + 1)
            self._driver.updatePV(reason)
            return True
        else:
            msg = '{}: could not execute'.format(reason)
            self._log_warning(msg)
            self._driver.setParam('ImgLog-Mon', msg)
            self._driver.updatePV('ImgLog-Mon')
            return False

    def _write_dvf_acquire(self, reason, value):
        """."""
        if reason != 'ImgDVFAcquire-Cmd':
            return None

        # set acquire
        if self.meas.set_acquire():
            value = self._driver.getParam(reason)
            self._driver.setParam(reason, value + 1)
            self._driver.updatePV(reason)
            return True
        else:
            msg = '{}: could not execute'.format(reason)
            self._log_warning(msg)
            self._driver.setParam('ImgLog-Mon', msg)
            self._driver.updatePV('ImgLog-Mon')
            return False

    def _log_warning(self, message):
        message = f'[{self.heartbeat}] ' + message
        _log.warning(message)

    def _write_fwhm_factor(self, reason, value):
        if reason not in (
                'ImgROIXUpdateWithFWHMFactor-SP',
                'ImgROIYUpdateWithFWHMFactor-SP'
        ):
            return None
        if 'X' in reason:
            self.meas.fwhmx_factor = value
        else:
            self.meas.fwhmy_factor = value
        self._write_pv_sp_rb(reason, value)

        return True

    def _write_update_roi_with_fwhm(self, reason, value):
        """."""
        if reason not in ('ImgROIUpdateWithFWHM-Sel'):
            return None
        self.meas.update_roi_with_fwhm = value
        self._write_pv_sp_rb(reason, value)

        return True

    def _conv_imgattr2value(self, attr):
        value = self.meas.image2dfit
        if value is None:
            # image2dfit object anomalously not created yet!
            return None
        if isinstance(attr, tuple):
            for obj in attr:
                value = getattr(value, obj)
        else:
            value = getattr(value, attr)
        return value



"""ID Search module."""

import copy as _copy

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..namesys import SiriusPVName as _SiriusPVName
from ..namesys import Filter as _Filter


class IDSearch:
    """ID Search Class."""

    # NOTE: Maybe we should move these mappings to
    # static tables in our cs-constants server.

    # NOTE: see
    # https://wiki-sirius.lnls.br/mediawiki/index.php/Machine:Insertion_Devices

    _beamline2idname = {
        'CARNAUBA': 'SI-06SB:ID-APU22',  # titular: VPU19
        'CATERETE': 'SI-07SP:ID-APU22',  # titular: VPU19
        'EMA':      'SI-08SB:ID-APU22',  # titular: IVU18
        'MANACA':   'SI-09SA:ID-APU22',  # titular: 2 x APU22
        'SABIA':    'SI-10SB:ID-DELTA52',  # titular: 2 x DELTA52 (EPU50 prev.)
        'IPE':      'SI-11SP:ID-APU58',  # titular: 2 x APPLE-II
        'PAINEIRA': 'SI-14SB:ID-WIG180',  # titular: IVU18
        'SAPUCAIA': 'SI-17SA:ID-PAPU50',  # titular: 2 x APU22
    }

    _idname2beamline = {v: k for k, v in _beamline2idname.items()}

    _idparam_fields = (
        'PERIOD_LENGTH',  # [mm]
        'KPARAM_MIN',  # [mm]
        'KPARAM_MAX',  # [mm]
        'KPARAM_PARKED',  # [mm]
        'KPARAM_POL_CHANGE',  # [mm]
        'KPARAM_TOL',  # [mm]
        'PPARAM_MIN',  # [mm]
        'PPARAM_MAX',  # [mm]
        'PPARAM_PARKED',  # [mm]
        'PPARAM_TOL',  # [mm]
        )

    _idname2params = {
        'SI-06SB:ID-APU22': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                22,
                0, 11, 11, 0, 0.01,
                None, None, None, None)),
        'SI-07SP:ID-APU22': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                22,
                0, 11, 11, 0, 0.01,
                None, None, None, None)),
        'SI-08SB:ID-APU22': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                22,
                0, 11, 11, 0, 0.01,
                None, None, None, None)),
        'SI-09SA:ID-APU22': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                22,
                0, 11, 11, 0, 0.01,
                None, None, None, None)),
        # NOTE: for EPU50 there is a large discrepancy
        # between RB/SP/Mon phase values
        'SI-10SB:ID-EPU50': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                50,
                +22, +300, +300, +300, 0.1,
                -25, 25, 0, 0.5)),
        'SI-10SB:ID-DELTA52': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                52.5,
                -52.5/2, +52.5/2, 0, 0, 0.020,
                -52.5/2, +52.5/2, 0, 0.010)),
        'SI-11SP:ID-APU58': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                58,
                0, 29, 29, 0, 0.01,
                None, None, None, None)),
        'SI-14SB:ID-WIG180': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                180,
                49.73, 49.73, 150, 150, 0.1,
                None, None, None, None)),
        'SI-17SA:ID-PAPU50': _get_namedtuple(
            'IDParameters',
            _idparam_fields, (
                50,
                0, 25, 25, 0, 0.1,
                None, None, None, None)),
    }

    POL_NONE_STR = 'none'
    POL_UNDEF_STR = 'undef'

    _idname2pol_sel = {
        'SI-06SB:ID-APU22': {
            0: ('horizontal', None),  # [mm]
        },
        'SI-07SP:ID-APU22': {
            0: ('horizontal', None),  # [mm]
        },
        'SI-08SB:ID-APU22': {
            0: ('horizontal', None),  # [mm]
        },
        'SI-09SA:ID-APU22': {
            0: ('horizontal', None),  # [mm]
        },
        'SI-10SB:ID-EPU50': {
            0: ('circularn', -16.36),  # [mm]
            1: ('horizontal', 0.00),  # [mm]
            2: ('circularp', +16.36),  # [mm]
            3: ('vertical', 25.00),  # [mm]
        },
        'SI-10SB:ID-DELTA52': {
            0: ('circularn', -52.5/4),  # [mm]
            1: ('horizontal', 0.00),  # [mm]
            2: ('circularp', +52.5/4),  # [mm]
            3: ('vertical', -52.5/2),  # [mm]
        },
        'SI-11SP:ID-APU58': {
            0: ('horizontal', None),  # [mm]
        },
        'SI-17SA:ID-PAPU50': {
            0: ('horizontal', None),  # [mm]
        },
    }
    _idname2pol_sts = _copy.deepcopy(_idname2pol_sel)
    _idname2pol_sts['SI-10SB:ID-EPU50'].update(
        {4: (POL_NONE_STR, None), 5: (POL_UNDEF_STR, None)})
    _idname2pol_sts['SI-10SB:ID-DELTA52'].update(
        {4: (POL_NONE_STR, None), 5: (POL_UNDEF_STR, None)})

    _idname_2_idff = {
        'SI-06SB:ID-APU22': None,
        'SI-07SP:ID-APU22': None,
        'SI-08SB:ID-APU22': None,
        'SI-09SA:ID-APU22': None,
        'SI-10SB:ID-EPU50': {
            'polarizations': tuple(
                item[0] for item in
                _idname2pol_sts['SI-10SB:ID-EPU50'].values()),
            'pparameter': 'SI-10SB:ID-EPU50:Phase-Mon',
            'kparameter': 'SI-10SB:ID-EPU50:Gap-Mon',
            'ch1': 'SI-10SB:PS-CH-1:Current-SP',  # upstream
            'ch2': 'SI-10SB:PS-CH-2:Current-SP',  # downstream
            'cv1': 'SI-10SB:PS-CV-1:Current-SP',
            'cv2': 'SI-10SB:PS-CV-2:Current-SP',
            'qs1': 'SI-10SB:PS-QS-1:Current-SP',
            'qs2': 'SI-10SB:PS-QS-2:Current-SP',
        },
        'SI-10SB:ID-DELTA52': {
            'polarizations': tuple(
                item[0] for item in
                _idname2pol_sts['SI-10SB:ID-DELTA52'].values()),
            'pparameter': 'SI-10SB:ID-DELTA52:PParam-Mon',
            'kparameter': 'SI-10SB:ID-DELTA52:KParam-Mon',
            'ch1': 'SI-10SB:PS-CH-1:Current-SP',  # upstream
            'ch2': 'SI-10SB:PS-CH-2:Current-SP',  # downstream
            'cv1': 'SI-10SB:PS-CV-1:Current-SP',
            'cv2': 'SI-10SB:PS-CV-2:Current-SP',
            'qs1': 'SI-10SB:PS-QS-1:Current-SP',
            'qs2': 'SI-10SB:PS-QS-2:Current-SP',
        },
        'SI-11SP:ID-APU58': None,
        'SI-14SB:ID-WIG180': {
            'polarizations': ('horizontal', ),
            'pparameter': None,
            'kparameter': 'SI-14SB:ID-WIG180:Gap-Mon',
            'ch1': 'SI-14SB:PS-CH-1:Current-SP',  # upstream
            'ch2': 'SI-14SB:PS-CH-2:Current-SP',  # downstream
        },
        'SI-17SA:ID-PAPU50': {
            'polarizations': ('horizontal', ),
            'pparameter': None,
            'kparameter': 'SI-17SA:ID-PAPU50:Phase-Mon',
            'ch1': 'SI-17SA:PS-CH-1:Current-SP',  # upstream
            'ch2': 'SI-17SA:PS-CH-2:Current-SP',  # downstream
            'cv1': 'SI-17SA:PS-CV-1:Current-SP',
            'cv2': 'SI-17SA:PS-CV-2:Current-SP',
        },
    }

    @staticmethod
    def get_idnames(filters=None):
        """Return a sorted and filtered list of all ID names."""
        idnames_list = list(IDSearch._idname_2_idff.keys())
        idnames = _Filter.process_filters(idnames_list, filters=filters)
        return sorted(idnames)

    @staticmethod
    def get_beamlines(filters=None):
        """Return a sorted and filtered list of all ID beamlines."""
        beamlines_list = list(IDSearch._idname2beamline.values())
        beamlines = _Filter.process_filters(beamlines_list, filters=filters)
        return sorted(beamlines)

    @staticmethod
    def conv_idname_2_beamline(idname):
        """Return the beam line name corresponding to sector."""
        if idname in IDSearch._idname2beamline.keys():
            return IDSearch._idname2beamline[idname]
        return None

    @staticmethod
    def conv_beamline_2_idname(beamline):
        """Return the beam line name corresponding to idname."""
        if beamline in IDSearch._beamline2idname.keys():
            return IDSearch._beamline2idname[beamline]
        return None

    @staticmethod
    def get_idname_2_beamline_dict():
        """Return idname to beamline dictionary."""
        return IDSearch._idname2beamline.copy()

    @staticmethod
    def get_beamline_2_idname_dict():
        """Return beamline to idname dictionary."""
        return IDSearch._beamline2idname.copy()

    @staticmethod
    def conv_idname_2_parameters(idname):
        """Return ID parameters from idname."""
        return IDSearch._idname2params[idname]

    @staticmethod
    def conv_idname_2_idff(idname):
        """Return the IDFF dictionary for a given ID name."""
        idff = IDSearch._idname_2_idff[idname]
        if idff is None:
            return None
        return _copy.deepcopy(idff)

    @staticmethod
    def conv_idname_2_pparameter_propty(idname):
        """."""
        idff = IDSearch.conv_idname_2_idff(idname)
        if idff is None:
            return None
        pparameter = idff['pparameter']
        if pparameter:
            pvname = _SiriusPVName(pparameter)
            return pvname.propty
        else:
            return None

    @staticmethod
    def conv_idname_2_kparameter_propty(idname):
        """."""
        idff = IDSearch.conv_idname_2_idff(idname)
        if idff is None:
            return None
        kparameter = idff['kparameter']
        if kparameter:
            pvname = _SiriusPVName(kparameter)
            return pvname.propty
        else:
            return None

    @staticmethod
    def conv_idname_2_idff_chnames(idname):
        """."""
        return IDSearch._get_devname_from_idff(idname, ('ch1', 'ch2'))

    @staticmethod
    def conv_idname_2_idff_cvnames(idname):
        """."""
        return IDSearch._get_devname_from_idff(idname, ('cv1', 'cv2'))

    @staticmethod
    def conv_idname_2_idff_qsnames(idname):
        """."""
        return IDSearch._get_devname_from_idff(idname, ('qs1', 'qs2'))

    @staticmethod
    def conv_idname_2_polarizations(idname):
        """Return ID polarizations (sel)."""
        pols = IDSearch._idname2pol_sel[idname]
        return tuple(pol[0] for pol in pols.values())

    @staticmethod
    def conv_idname_2_polarizations_sts(idname):
        """Return ID polarizations (sts)."""
        pols = IDSearch._idname2pol_sts[idname]
        return tuple(pol[0] for pol in pols.values())

    @staticmethod
    def conv_idname_2_polarization_state(idname, pparameter, kparameter):
        """Return polarization state index."""
        params = IDSearch.conv_idname_2_parameters(idname)
        pols_sts = IDSearch._idname2pol_sts[idname]

        # check if polarization is defined
        for pol_idx, pol in pols_sts.items():
            _, pol_phase = pol
            if pol_phase is None:  # ignore none and undef configs
                continue
            if abs(pparameter - pol_phase) <= params.PPARAM_TOL:
                return pol_idx

        pol_state_sts = IDSearch.conv_idname_2_polarizations_sts(idname)

        # checking if changing polarization
        if abs(kparameter - params.KPARAM_POL_CHANGE) <= params.KPARAM_TOL:
            pol_idx = pol_state_sts.index(IDSearch.POL_NONE_STR)
            return pol_idx

        # at this point the configuration must be undefined
        pol_idx = pol_state_sts.index(IDSearch.POL_UNDEF_STR)
        return pol_idx

    @staticmethod
    def conv_idname_2_polarization_pparameter(idname, pol):
        """Return pparameter value of a given polarization."""
        pols = IDSearch._idname2pol_sel[idname]
        if isinstance(pol, int):
            return pols[pol][1]
        elif isinstance(pol, str):
            for pol_name, pol_pparam in pols.values():
                if pol == pol_name:
                    return pol_pparam
            raise ValueError(f'Invalid polarization string "{pol}"')
        else:
            raise TypeError('Invalid polarization type.')

    # --- private ----

    @staticmethod
    def _get_devname_from_idff(idname, correctors):
        idff = IDSearch.conv_idname_2_idff(idname)
        corrs = list()
        for corr in correctors:
            if corr in idff:
                pvname = _SiriusPVName(idff[corr])
                corrs.append(pvname.device_name)
        return corrs



"""Beam Loss Monitor devices."""

from ..util import ClassProperty as _classproperty
from ..search import BLMSearch as _BLMSearch
from .device import Device as _Device


class BLM(_Device):
    """Beam Loss Monitor Device."""

    class DEVICES:
        """Devices names."""
        __SENSOR1 = __SENSOR2 = __SENSOR3 = __SENSOR4 = None
        __SENSOR5 = __SENSOR6 = __SENSOR7 = __SENSOR8 = None

        @_classproperty
        def SENSOR1(cls):
            """."""
            if cls.__SENSOR1 is None:
                cls.__SENSOR1 = _BLMSearch.conv_blmname_2_devname('SENSOR1')
            return cls.__SENSOR1

        @_classproperty
        def SENSOR2(cls):
            """."""
            if cls.__SENSOR2 is None:
                cls.__SENSOR2 = _BLMSearch.conv_blmname_2_devname('SENSOR2')
            return cls.__SENSOR2

        @_classproperty
        def SENSOR3(cls):
            """."""
            if cls.__SENSOR3 is None:
                cls.__SENSOR3 = _BLMSearch.conv_blmname_2_devname('SENSOR3')
            return cls.__SENSOR3

        @_classproperty
        def SENSOR4(cls):
            """."""
            if cls.__SENSOR4 is None:
                cls.__SENSOR4 = _BLMSearch.conv_blmname_2_devname('SENSOR4')
            return cls.__SENSOR4

        @_classproperty
        def SENSOR5(cls):
            """."""
            if cls.__SENSOR5 is None:
                cls.__SENSOR5 = _BLMSearch.conv_blmname_2_devname('SENSOR5')
            return cls.__SENSOR5

        @_classproperty
        def SENSOR6(cls):
            """."""
            if cls.__SENSOR6 is None:
                cls.__SENSOR6 = _BLMSearch.conv_blmname_2_devname('SENSOR6')
            return cls.__SENSOR6

        @_classproperty
        def SENSOR7(cls):
            """."""
            if cls.__SENSOR7 is None:
                cls.__SENSOR7 = _BLMSearch.conv_blmname_2_devname('SENSOR7')
            return cls.__SENSOR7

        @_classproperty
        def SENSOR8(cls):
            """."""
            if cls.__SENSOR8 is None:
                cls.__SENSOR8 = _BLMSearch.conv_blmname_2_devname('SENSOR8')
            return cls.__SENSOR8

        @_classproperty
        def ALL(cls):
            return (
                cls.SENSOR1, cls.SENSOR2, cls.SENSOR3, cls.SENSOR4,
                cls.SENSOR5, cls.SENSOR6, cls.SENSOR7, cls.SENSOR8)

    PROPERTIES_DEFAULT = (
        'Count-Mon', )

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in BLM.DEVICES.ALL:
            raise NotImplementedError(devname)
        self._blmname = _BLMSearch.conv_devname_2_blmname(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def blmname(self):
        """Return BLM tag name."""
        return self._blmname

    @property
    def count_mon(self):
        """Return BLM count value [pulse/s]."""
        return self['Count-Mon']

    @property
    def secsub(self):
        """Return sector-subsector where BLM is located."""
        return _BLMSearch.conv_blmname_2_secsub(self.blmname)

    @property
    def position(self):
        """Return installation position of the BLM."""
        return _BLMSearch.conv_blmname_2_position(self.blmname)

    @property
    def channel(self):
        """Return counter channel of BLM."""
        return _BLMSearch.conv_blmname_2_counter_channel(self.blmname)



"""Optics correction module."""


import numpy as _np
from ..clientconfigdb import ConfigDBDocument as _ConfigDBDocument
from .csdev import Const as _Const


class OpticsCorr:
    """Main class for optics correction.

    Calculate Deltas in Quadrupoles and Sextupoles Integrated Strengths.
    Store all correction parameters.
    """

    def __init__(self, magnetfams_ordering,
                 nominal_matrix, nominal_intstrengths, nominal_opticsparam,
                 magnetfams_focusing, magnetfams_defocusing):
        """Class constructor."""
        # declaration of attributes (as pylint requires)
        self._nominal_matrix = None
        self._initialized = False
        self._nominal_intstrengths = None
        self._matrix_prop_svd = None
        self._matrix_prop_2knobs = None
        self._inverse_matrix_prop_svd = None
        self._inverse_matrix_prop_2knobs = None
        self._nominal_opticsparam = None
        self._magnetfams_focusing = None
        self._magnetfams_defocusing = None

        # initialization of attributes
        self._set_magnetfams_ordering(magnetfams_ordering)
        self.nominal_matrix = nominal_matrix
        self.nominal_intstrengths = nominal_intstrengths
        self.nominal_opticsparam = nominal_opticsparam
        self.magnetfams_focusing = magnetfams_focusing
        self.magnetfams_defocusing = magnetfams_defocusing

        self._calculate_matrices()
        self._initialized = True

    @property
    def magnetfams_ordering(self):
        """List of strings of the order of magnet families used on correction.

        Correction matrix and nominal integrated strengths must be coherent to
        this property.
        """
        return self._magnetfams_ordering

    def _set_magnetfams_ordering(self, value):
        """Set magnetfams_ordering_svd property."""
        if not isinstance(value, tuple):
            raise TypeError("Value must be a tuple.")
        for item in value:
            if not isinstance(item, str):
                raise TypeError("List elements must be strings.")
        if len(value) < 2:
            raise ValueError("At least 2 magnet families are necessary to "
                             "correction!")

        self._magnetfams_ordering = value

    @property
    def nominal_matrix(self):
        """Nominal matrix, correspondent to add method and multiple knobs.

        Considers all knobs listed in 'magnetfams_ordering' property.
        """
        return self._nominal_matrix

    @nominal_matrix.setter
    def nominal_matrix(self, value):
        if not isinstance(value, list):
            raise TypeError("Value must be a flat list (C-like format.")
        for item in value:
            if not isinstance(item, float):
                raise TypeError("List elements must be floats.")

        matrix = _np.array(value)
        matrix = _np.reshape(matrix, [2, len(self._magnetfams_ordering)])

        self._nominal_matrix = matrix

        if self._initialized:
            self._calculate_matrices()

    @property
    def nominal_intstrengths(self):
        """Nominal integrated strengths."""
        return self._nominal_intstrengths

    @nominal_intstrengths.setter
    def nominal_intstrengths(self, value):
        if not isinstance(value, list):
            raise TypeError("Value must be a flat list (C-like format.")
        for item in value:
            if not isinstance(item, float):
                raise TypeError("List elements must be floats.")

        intstrengths = _np.array(value)
        intstrengths = _np.reshape(intstrengths,
                                   [1, len(self._magnetfams_ordering)])

        self._nominal_intstrengths = intstrengths

        if self._initialized:
            self._matrix_prop_svd = self._calculate_matrix_svd(method=0)
            self._matrix_prop_2knobs = self._calculate_matrix_2knobs(method=0)
            self._inverse_matrix_prop_svd = (
                self._calculate_inverse(method=0, grouping='svd'))
            self._inverse_matrix_prop_2knobs = (
                self._calculate_inverse(method=0, grouping='2knobs'))

    @property
    def nominal_opticsparam(self):
        """Nominal optics parameter."""
        return self._nominal_opticsparam

    @nominal_opticsparam.setter
    def nominal_opticsparam(self, value):
        if not isinstance(value, list):
            raise TypeError("Value must be a flat list (C-like format.")
        for item in value:
            if not isinstance(item, float):
                raise TypeError("List elements must be floats.")

        opticsparam = _np.array(value)
        opticsparam = _np.reshape(opticsparam, [2, 1])

        self._nominal_opticsparam = opticsparam

    @property
    def magnetfams_focusing(self):
        """List of strings of focusing magnet families used on correction.

        It corresponds to the first correction knob magnet families.
        """
        return self._magnetfams_focusing

    @magnetfams_focusing.setter
    def magnetfams_focusing(self, value):
        if not isinstance(value, tuple):
            raise TypeError("Value must be a tuple.")
        for item in value:
            if not isinstance(item, str):
                raise TypeError("List elements must be strings.")
        if not value:
            raise ValueError("At least one focusing magnet family is "
                             "necessary to correction!")
        if not all(item in self._magnetfams_ordering for item in value):
            raise ValueError("Focusing magnet families must be part of the "
                             "'magnetfams_ordering' property!")

        self._magnetfams_focusing = tuple(sorted(
            value, key=self.magnetfams_ordering.index))

        if self._initialized:
            self._calculate_matrices()

    @property
    def magnetfams_defocusing(self):
        """List of strings of defocusing magnet families used on correction.

        It corresponds to the second correction knob magnet families.
        """
        return self._magnetfams_defocusing

    @magnetfams_defocusing.setter
    def magnetfams_defocusing(self, value):
        if not isinstance(value, tuple):
            raise TypeError("Value must be a tuple.")
        for item in value:
            if not isinstance(item, str):
                raise TypeError("List elements must be strings.")
        if not value:
            raise ValueError("At least one defocusing magnet family is "
                             "necessary to correction!")
        if not all(item in self._magnetfams_ordering for item in value):
            raise ValueError("Defocusing magnet families must be part of the "
                             "'magnetfams_ordering_svd' property!")

        self._magnetfams_defocusing = tuple(sorted(
            value, key=self.magnetfams_ordering.index))

        if self._initialized:
            self._calculate_matrices()

    @property
    def matrix_add_svd(self):
        """Matrix of additional method and multiple knobs."""
        return self._matrix_add_svd

    @property
    def matrix_prop_svd(self):
        """Matrix of proportional method and multiple knobs."""
        return self._matrix_prop_svd

    @property
    def matrix_add_2knobs(self):
        """Matrix of additional method and 2 knobs."""
        return self._matrix_add_2knobs

    @property
    def matrix_prop_2knobs(self):
        """Matrix of proportional method and 2 knobs."""
        return self._matrix_prop_2knobs

    @property
    def inverse_matrix_add_svd(self):
        """Inverse matrix of additional method and multiple knobs."""
        return self._inverse_matrix_add_svd

    @property
    def inverse_matrix_prop_svd(self):
        """Inverse matrix of proportional method and multiple knobs."""
        return self._inverse_matrix_prop_svd

    @property
    def inverse_matrix_add_2knobs(self):
        """Inverse matrix of additional method and 2 knobs."""
        return self._inverse_matrix_add_2knobs

    @property
    def inverse_matrix_prop_2knobs(self):
        """Inverse matrix of proportional method and 2 knobs."""
        return self._inverse_matrix_prop_2knobs

    def calculate_delta_intstrengths(self, method, grouping,
                                     delta_opticsparam):
        """Calculate the delta on integrated strengths.

        Based on the required optics parameter delta (tune or chromaticity).

        Return a flat numpy.array of deltas of all magnet families of
        'magnetfams_ordering' property. Deltas will only be calculated for
        families listed on 'magnetfams_focusing' and 'magnetfams_defocusing'
        properties. The rest of the families will have null deltas.
        """
        inv = self._choose_inverse_matrix(method, grouping)
        delta_opticsparam = _np.array([[delta_opticsparam[0]],
                                       [delta_opticsparam[1]]])
        magnetfams_delta_intstrengths = _np.zeros(
            len(self.magnetfams_ordering))

        delta = _np.dot(inv, delta_opticsparam).T.flatten()

        foc = len(self.magnetfams_focusing)
        defoc = len(self.magnetfams_defocusing)
        if method == 0:
            if grouping == 'svd':
                for i in range(foc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_focusing[i])
                    magnetfams_delta_intstrengths[index] = (
                        self.nominal_intstrengths[0, index] * delta[i])
                for i in range(defoc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_defocusing[i])
                    magnetfams_delta_intstrengths[index] = (
                        self.nominal_intstrengths[0, index] * delta[i+foc])
            elif grouping == '2knobs':
                for i in range(foc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_focusing[i])
                    magnetfams_delta_intstrengths[index] = (
                        self.nominal_intstrengths[0, index] * delta[0])
                for i in range(defoc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_defocusing[i])
                    magnetfams_delta_intstrengths[index] = (
                        self.nominal_intstrengths[0, index] * delta[1])
        elif method == 1:
            if grouping == 'svd':
                for i in range(foc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_focusing[i])
                    magnetfams_delta_intstrengths[index] = delta[i]
                for i in range(defoc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_defocusing[i])
                    magnetfams_delta_intstrengths[index] = delta[i+foc]
            elif grouping == '2knobs':
                for i in range(foc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_focusing[i])
                    magnetfams_delta_intstrengths[index] = delta[0]
                for i in range(defoc):
                    index = self.magnetfams_ordering.index(
                        self.magnetfams_defocusing[i])
                    magnetfams_delta_intstrengths[index] = delta[1]

        return magnetfams_delta_intstrengths.flatten()

    def calculate_opticsparam(self, delta_intstrengths):
        """Calculate the optics parameter (tune or chromaticity).

        Based on the required integrated strengths.
        """
        delta_intstrengths = _np.array([delta_intstrengths]).transpose()

        delta_opticsparam = _np.dot(self.nominal_matrix, delta_intstrengths)
        opticsparam = self.nominal_opticsparam + delta_opticsparam

        return opticsparam.flatten()

    # Private methods

    def _calculate_matrices(self):
        """Recalculate all matrices and their ineverse matrices."""
        self._matrix_add_svd = self._calculate_matrix_svd(method=1)
        self._matrix_prop_svd = self._calculate_matrix_svd(method=0)
        self._matrix_add_2knobs = self._calculate_matrix_2knobs(method=1)
        self._matrix_prop_2knobs = self._calculate_matrix_2knobs(method=0)
        self._inverse_matrix_add_svd = self._calculate_inverse(
            method=1, grouping='svd')
        self._inverse_matrix_prop_svd = self._calculate_inverse(
            method=0, grouping='svd')
        self._inverse_matrix_add_2knobs = self._calculate_inverse(
            method=1, grouping='2knobs')
        self._inverse_matrix_prop_2knobs = self._calculate_inverse(
            method=0, grouping='2knobs')

    def _calculate_matrix_svd(self, method):
        """Calculate matrix of multiple knobs (svd)."""
        if method == 0:  # proportional method
            matrix = self.nominal_matrix*self.nominal_intstrengths
        elif method == 1:  # additional method
            matrix = self.nominal_matrix

        foc = len(self.magnetfams_focusing)
        defoc = len(self.magnetfams_defocusing)
        mat_svd = _np.zeros([2, foc+defoc])

        for i in range(foc):
            idxc = self.magnetfams_ordering.index(self.magnetfams_focusing[i])
            mat_svd[0, i] = matrix[0, idxc]
            mat_svd[1, i] = matrix[1, idxc]
        for i in range(defoc):
            idxc = self.magnetfams_ordering.index(
                self.magnetfams_defocusing[i])
            mat_svd[0, i+foc] = matrix[0, idxc]
            mat_svd[1, i+foc] = matrix[1, idxc]

        return mat_svd

    def _calculate_matrix_2knobs(self, method):
        """Calculate matrices of 2 knobs."""
        if method == 0:  # proportional method
            matrix = self.nominal_matrix*self.nominal_intstrengths
        elif method == 1:  # additional method
            matrix = self.nominal_matrix

        mat_2knobs = _np.array([[0.0, 0.0], [0.0, 0.0]])
        # matrix indices are organized like:
        # [focusing magnet fams, X plan    defocusing magnet fams, X plan]
        # [focusing magnet fams, Y plan    defocusing magnet fams, Y plan]

        for fam in self.magnetfams_focusing:
            index = self.magnetfams_ordering.index(fam)
            mat_2knobs[0, 0] += matrix[0, index]
            mat_2knobs[1, 0] += matrix[1, index]

        for fam in self.magnetfams_defocusing:
            index = self.magnetfams_ordering.index(fam)
            mat_2knobs[0, 1] += matrix[0, index]
            mat_2knobs[1, 1] += matrix[1, index]

        return mat_2knobs

    def _calculate_inverse(self, method, grouping):
        """Calculate inverse of a matrix using SVD."""
        matrix = self._choose_matrix(method, grouping)

        try:
            umat, smat, vmat = _np.linalg.svd(matrix, full_matrices=False)
        except _np.linalg.LinAlgError():
            raise Exception("Could not calculate SVD.")
        inv = _np.dot(_np.dot(vmat.T, _np.diag(1/smat)), umat.T)
        isnan = _np.any(_np.isnan(inv))
        isinf = _np.any(_np.isinf(inv))
        if isnan or isinf:
            raise Exception("Pseudo inverse contains nan or inf.")

        return inv

    def _choose_matrix(self, method, grouping):
        if method == 0:
            if grouping == 'svd':
                matrix = self.matrix_prop_svd
            elif grouping == '2knobs':
                matrix = self.matrix_prop_2knobs
        elif method == 1:
            if grouping == 'svd':
                matrix = self.matrix_add_svd
            elif grouping == '2knobs':
                matrix = self.matrix_add_2knobs
        return matrix

    def _choose_inverse_matrix(self, method, grouping):
        if method == 0:
            if grouping == 'svd':
                matrix = self.inverse_matrix_prop_svd
            elif grouping == '2knobs':
                matrix = self.inverse_matrix_prop_2knobs
        elif method == 1:
            if grouping == 'svd':
                matrix = self.inverse_matrix_add_svd
            elif grouping == '2knobs':
                matrix = self.inverse_matrix_add_2knobs
        return matrix


class BOTuneCorr(OpticsCorr, _ConfigDBDocument):
    """Auxiliar class to handle Booster tune correction."""

    def __init__(self, name):
        """Get configuration from ConfigServer and initialize object."""
        _ConfigDBDocument.__init__(self, 'bo_tunecorr_params', name=name)

        self.load()
        params = self.value
        nominal_matrix = [item for sublist in params['matrix']
                          for item in sublist]
        nominal_kl = params['nominal KLs']
        OpticsCorr.__init__(
            self,
            magnetfams_ordering=_Const.BO_QFAMS_TUNECORR,
            nominal_matrix=nominal_matrix,
            nominal_intstrengths=nominal_kl,
            nominal_opticsparam=[0.0, 0.0],
            magnetfams_focusing=('QF',),
            magnetfams_defocusing=('QD',))

    def calculate_deltaKL(self, delta_tune):
        """Calculate delta KL based on the required delta tune."""
        return self.calculate_delta_intstrengths(
            method=1, grouping='2knobs', delta_opticsparam=delta_tune)

    def calculate_deltaTune(self, deltaKL):
        """Calculate delta tune based on the required delta KL."""
        return self.calculate_opticsparam(delta_intstrengths=deltaKL)


class BOChromCorr(OpticsCorr, _ConfigDBDocument):
    """Auxiliar class to handle Booster chromacity correction."""

    def __init__(self, name):
        """Get configuration from ConfigServer and initialize object."""
        _ConfigDBDocument.__init__(self, 'bo_chromcorr_params', name=name)

        self.load()
        params = self.value
        nominal_matrix = [item for sublist in params['matrix']
                          for item in sublist]
        nominal_sl = params['nominal SLs']
        nominal_chrom = params['nominal chrom']
        OpticsCorr.__init__(self,
                            magnetfams_ordering=_Const.BO_SFAMS_CHROMCORR,
                            nominal_matrix=nominal_matrix,
                            nominal_intstrengths=nominal_sl,
                            nominal_opticsparam=nominal_chrom,
                            magnetfams_focusing=('SF',),
                            magnetfams_defocusing=('SD',))

    def calculate_deltaSL(self, delta_chrom):
        """Calculate delta SL based on the required delta chromacity."""
        return self.calculate_delta_intstrengths(
            method=1, grouping='2knobs', delta_opticsparam=delta_chrom)

    def calculate_Chrom(self, deltaSL):
        """Calculate delta chromacity based on the required delta SL."""
        return self.calculate_opticsparam(delta_intstrengths=deltaSL)



#!/usr/bin/env python-sirius

"""Unittest module for hl_time_search.py."""

from unittest import TestCase
from siriuspy import util
from siriuspy.search import hl_time_search

mock_flag = True

public_interface = ('HLTimeSearch', )


class TestModule(TestCase):
    """Test module interface."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
                hl_time_search,
                public_interface)
        self.assertTrue(valid)


class TestHLTimeSearch(TestCase):
    """Test HLTimeSearch class."""

    public_interface = (
        'get_hl_events',
        'get_hl_triggers',
        'get_hl_trigger_database',
        'get_hl_trigger_sources',
        'get_hl_trigger_prop_value',
        'get_hl_trigger_prop_limits',
        'get_hl_trigger_interface',
        'get_ll_trigger_names',
        'get_hl_from_ll_triggers',
        'is_digital_input',
        'has_delay_type',
        'has_direction',
        'has_clock',
        'has_log',
        'check_hl_triggers_consistency',
        'reset',
    )

    def get_hl_triggers(self):
        """Test get_hl_triggers."""
        # TODO: implement test!
        pass

    def get_hl_trigger_database(self):
        """Test get_hl_trigger_database."""
        # TODO: implement test!
        pass

    def get_hl_trigger_sources(self):
        """Test get_hl_trigger_sources."""
        # TODO: implement test!
        pass

    def get_hl_trigger_prop_value(self):
        """Test get_hl_trigger_prop_value."""
        # TODO: implement test!
        pass

    def get_hl_trigger_prop_limits(self):
        """Test get_hl_trigger_prop_limits."""
        # TODO: implement test!
        pass

    def get_hl_trigger_interface(self):
        """Test get_hl_trigger_interface."""
        # TODO: implement test!
        pass

    def get_ll_trigger_names(self):
        """Test get_ll_trigger_names."""
        # TODO: implement test!
        pass

    def get_hl_from_triggers(self):
        """Test get_hl_from_ll_triggers."""
        # TODO: implement test!
        pass

    def has_delay_type(self):
        """Test has_delay_type."""
        # TODO: implement test!
        pass

    def has_clock(self):
        """Test has_clock."""
        # TODO: implement test!
        pass

    def check_hl_triggers_consistency(self):
        """Test check_hl_triggers_consistency."""
        # TODO: implement test!
        pass



#!/usr/bin/env python-sirius

"""Unittest module for ll_time_search.py."""

from unittest import TestCase
import re
from siriuspy import util
from siriuspy.search import ll_time_search

mock_flag = True

public_interface = (
    'LLTimeSearch',
)


class TestModule(TestCase):
    """Test module interface."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
                ll_time_search,
                public_interface)
        self.assertTrue(valid)


class TestLLTimeSearch(TestCase):
    """Test LLTimeSearch class."""

    public_interface = (
        'LLRegExp',
        'In2OutMap',
        'Out2InMap',
        'get_channel_input',
        'get_channel_output_port_pvname',
        'get_channel_internal_trigger_pvname',
        'get_device_names',
        'get_evg_name',
        'get_evg2fout_mapping',
        'get_triggersource_devices',
        'get_trigsrc2fout_mapping',
        'get_fout2trigsrc_mapping',
        'get_device_tree',
        'reset',
        'server_online',
        'get_connections_from_evg',
        'get_connections_twds_evg',
        'get_top_chain_senders',
        'get_final_receivers',
        'get_relations_from_evg',
        'get_relations_twds_evg',
        'is_digital_input',
        'has_delay_type',
        'has_direction',
        'has_clock',
        'has_log',
        'get_trigger_name',
        'get_fout_channel',
        'get_evg_channel',
        'get_crates_mapping',
    )

    def test_public_interface(self):
        """Test class public interface."""
        valid = util.check_public_interface_namespace(
            ll_time_search.LLTimeSearch, TestLLTimeSearch.public_interface)
        self.assertTrue(valid)

    def test_LLRegExp(self):
        """Test LLRegExp."""
        typ = type(re.compile('dummy_input'))
        self.assertIsInstance(ll_time_search.LLTimeSearch.LLRegExp, typ)

    def test_In2OutMap(self):
        """Test In2OutMap."""
        self.assertIsInstance(ll_time_search.LLTimeSearch.In2OutMap, dict)

    def test_Out2InMap(self):
        """Test Out2InMap."""
        self.assertIsInstance(ll_time_search.LLTimeSearch.Out2InMap, dict)

    def test_get_channel_input(self):
        """Test get_channel_input."""
        # TODO: implement test!
        pass

    def test_add_crates_info(self):
        """Test add_crates_info."""
        # TODO: implement test!
        pass

    def test_add_bbb_info(self):
        """Test add_bbb_info."""
        # TODO: implement test!
        pass

    def test_get_device_names(self):
        """Test get_devices."""
        # TODO: implement test!
        pass

    def test_get_device_tree(self):
        """Test get_device_tree."""
        # TODO: implement test!
        pass

    def test_reset(self):
        """Test reset."""
        # TODO: implement test!
        pass

    def test_server_online(self):
        """Test server_online."""
        # TODO: implement test!
        pass

    def test_get_connections_from_evg(self):
        """Test get_connections_from_evg."""
        # TODO: implement test!
        pass

    def test_get_connections_twds_evg(self):
        """Test get_connections_twds_evg."""
        # TODO: implement test!
        pass

    def test_get_top_chain_senders(self):
        """Test get_top_chain_senders."""
        # TODO: implement test!
        pass

    def test_get_final_receivers(self):
        """Test get_final_receivers."""
        # TODO: implement test!
        pass

    def test_get_relations_from_evg(self):
        """Test get_relations_from_evg."""
        # TODO: implement test!
        pass

    def test_get_relations_twds_evg(self):
        """Test get_relations_twds_evg."""
        # TODO: implement test!
        pass

    def test_get_hierarchy_list(self):
        """Test get_hierarchy_list."""
        # TODO: implement test!
        pass



"""Sirius PV class."""


import epics as _epics


class PV(_epics.pv.PV):
    """PV class."""

    def set_auto_monitor(self, value):
        """Set auto_monitor property."""
        self.auto_monitor = value



"""."""

from .device import Device as _Device


class ICT(_Device):
    """ICT Device."""

    class DEVICES:
        """Devices names."""

        LI_1 = 'LI-01:DI-ICT-1'
        LI_2 = 'LI-01:DI-ICT-2'
        TB_02 = 'TB-02:DI-ICT'
        TB_04 = 'TB-04:DI-ICT'
        TS_01 = 'TS-01:DI-ICT'
        TS_04 = 'TS-04:DI-ICT'
        ALL = (LI_1, LI_2, TB_02, TB_04, TS_01, TS_04, )

    PROPERTIES_DEFAULT = (
        'Charge-Mon', 'ChargeAvg-Mon', 'ChargeMax-Mon',
        'ChargeMin-Mon', 'ChargeStd-Mon', 'PulseCount-Mon')

    def __init__(self, devname, props2init='all'):
        """."""
        # check if device exists
        if devname not in ICT.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, props2init=props2init)

    @property
    def charge(self):
        """."""
        return self['Charge-Mon']

    @property
    def charge_avg(self):
        """."""
        return self['ChargeAvg-Mon']

    @property
    def charge_max(self):
        """."""
        return self['ChargeMax-Mon']

    @property
    def charge_min(self):
        """."""
        return self['ChargeMin-Mon']

    @property
    def charge_std(self):
        """."""
        return self['ChargeStd-Mon']

    @property
    def pulse_count(self):
        """."""
        return self['PulseCount-Mon']


class TranspEff(_Device):
    """Linac Transport Efficiency Device."""

    class DEVICES:
        """Devices names."""

        LI = 'LI-Glob:AP-TranspEff'
        ALL = (LI, )

    PROPERTIES_DEFAULT = ('Eff-Mon', 'EffAvg-Mon', )

    def __init__(self, devname, props2init='all'):
        """."""
        # check if device exists
        if devname not in TranspEff.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def efficiency(self):
        """."""
        return self['Eff-Mon']

    @property
    def efficiency_avg(self):
        """."""
        return self['EffAvg-Mon']



#!/usr/local/bin/python-sirius
"""RF Diag PVs."""

import numpy as _np

from .csdev import Const as _Const


class BORFStatusPV:
    """BO RF Status PV."""

    BIT_DISCONN = 0b00001
    BIT_SIINTLK = 0b00010
    BIT_LLINTLK = 0b00100
    BIT_RMPDSBL = 0b01000
    BIT_RMPNRDY = 0b10000

    PV_SIRIUS_INTLK = 0
    PV_LLRF_INTLK = 1
    PV_RMP_ENBLD = 2
    PV_RMP_READY = 3

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute Status PV."""
        value = 0

        # connected?
        disconnected = \
            not computed_pv.pvs[BORFStatusPV.PV_SIRIUS_INTLK].connected or \
            not computed_pv.pvs[BORFStatusPV.PV_LLRF_INTLK].connected or \
            not computed_pv.pvs[BORFStatusPV.PV_RMP_ENBLD].connected or \
            not computed_pv.pvs[BORFStatusPV.PV_RMP_READY].connected

        if disconnected:
            value |= BORFStatusPV.BIT_DISCONN
            value |= BORFStatusPV.BIT_SIINTLK
            value |= BORFStatusPV.BIT_LLINTLK
            value |= BORFStatusPV.BIT_RMPDSBL
            value |= BORFStatusPV.BIT_RMPNRDY
            return {'value': value}

        # sirius interlock?
        sts = computed_pv.pvs[BORFStatusPV.PV_SIRIUS_INTLK].value
        if sts != 0 or sts is None:
            value |= BORFStatusPV.BIT_SIINTLK

        # llrf interlock?
        sts = computed_pv.pvs[BORFStatusPV.PV_LLRF_INTLK].value
        if sts != 0 or sts is None:
            value |= BORFStatusPV.BIT_LLINTLK

        # rmp enabled?
        sts = computed_pv.pvs[BORFStatusPV.PV_RMP_ENBLD].value
        if sts != _Const.DsblEnbl.Enbl or sts is None:
            value |= BORFStatusPV.BIT_RMPDSBL

        # rmp ready?
        sts = computed_pv.pvs[BORFStatusPV.PV_RMP_READY].value
        if sts != _Const.DsblEnbl.Enbl or sts is None:
            value |= BORFStatusPV.BIT_RMPNRDY

        return {'value': value}


class CheckErrPV:
    """Check error PV and filter spurious spikes."""

    PV_ERR = 0
    _tol = 0

    def __init__(self):
        self._hist = [0]*10
        self._idx = 0

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Check error exceeds tolerance."""
        if not computed_pv.pvs[CheckErrPV.PV_ERR].connected:
            return None
        value = computed_pv.pvs[CheckErrPV.PV_ERR].value
        self._hist[self._idx] = value
        self._idx += 1
        self._idx %= 10
        sts = 1 if _np.all([v > self._tol for v in self._hist]) else 0
        return {'value': sts}


class SICheckAmpErrPV(CheckErrPV):
    """Check amplitude error."""

    _tol = _Const.SI_SL_ERRTOL_AMP


class SICheckPhsErrPV(CheckErrPV):
    """Check phase error."""

    _tol = _Const.SI_SL_ERRTOL_PHS


class SICheckDTuneErrPV(CheckErrPV):
    """Check detune error."""

    _tol = _Const.SI_SL_ERRTOL_DTU


class SIRFStatusPV:
    """SI RF Status PV."""

    BIT_DISCONN = 0b000001
    BIT_SIINTLK = 0b000010
    BIT_LLINTLK = 0b000100
    BIT_AMPLERR = 0b001000
    BIT_PHSEERR = 0b010000
    BIT_DTUNERR = 0b100000

    PV_SIRIUS_INTLK = 0
    PV_LLRF_INTLK = 1
    PV_AMPL_ERR = 2
    PV_PHSE_ERR = 3
    PV_DTUN_ERR = 4

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute Status PV."""
        value = 0

        # connected?
        disconnected = \
            not computed_pv.pvs[SIRFStatusPV.PV_SIRIUS_INTLK].connected or \
            not computed_pv.pvs[SIRFStatusPV.PV_LLRF_INTLK].connected or \
            not computed_pv.pvs[SIRFStatusPV.PV_AMPL_ERR].connected or \
            not computed_pv.pvs[SIRFStatusPV.PV_PHSE_ERR].connected or \
            not computed_pv.pvs[SIRFStatusPV.PV_DTUN_ERR].connected
        if disconnected:
            value |= SIRFStatusPV.BIT_DISCONN
            value |= SIRFStatusPV.BIT_SIINTLK
            value |= SIRFStatusPV.BIT_LLINTLK
            value |= SIRFStatusPV.BIT_AMPLERR
            value |= SIRFStatusPV.BIT_PHSEERR
            value |= SIRFStatusPV.BIT_DTUNERR
            return {'value': value}

        # sirius interlock?
        sts = computed_pv.pvs[SIRFStatusPV.PV_SIRIUS_INTLK].value
        if sts != 0 or sts is None:
            value |= SIRFStatusPV.BIT_SIINTLK

        # llrf interlock?
        sts = computed_pv.pvs[SIRFStatusPV.PV_LLRF_INTLK].value
        if sts != 0 or sts is None:
            value |= SIRFStatusPV.BIT_LLINTLK

        # amplitude error?
        severity = computed_pv.pvs[SIRFStatusPV.PV_AMPL_ERR].severity
        if severity != 0:
            value |= SIRFStatusPV.BIT_AMPLERR

        # phase error?
        severity = computed_pv.pvs[SIRFStatusPV.PV_PHSE_ERR].severity
        if severity != 0:
            value |= SIRFStatusPV.BIT_PHSEERR

        # detune error?
        severity = computed_pv.pvs[SIRFStatusPV.PV_DTUN_ERR].severity
        if severity != 0:
            value |= SIRFStatusPV.BIT_DTUNERR

        return {'value': value}



"""Insertion Device Feedforward Configuration."""

from copy import deepcopy as _dcopy

import numpy as _np

from ..search import IDSearch as _IDSearch
from ..clientconfigdb import ConfigDBDocument as _ConfigDBDocument


class IDFFConfig(_ConfigDBDocument):
    """Insertion Device Feedforward Configuration."""

    CONFIGDB_TYPE = 'si_idff'

    def __init__(self, name=None, url=None):
        """."""
        name_ = name or 'idff_' + self.generate_config_name()
        self._idname = None
        self._polarization_definitions = None
        super().__init__(
            config_type=IDFFConfig.CONFIGDB_TYPE, name=name_, url=url)

    @property
    def idname(self):
        """Return idname corresponding to IDFFConfig."""
        return self._idname

    @idname.setter
    def idname(self, value):
        """Set idname."""
        if value not in _IDSearch.get_idnames():
            raise ValueError(f'{value} is not a valid idname!')
        self._idname = value

    @property
    def pparameter_pvname(self):
        """Return ID pparameter pvname."""
        config = self._value
        if config:
            kparm = config['pvnames']['pparameter']
            return kparm
        else:
            raise ValueError('Configuration not defined!')

    @property
    def kparameter_pvname(self):
        """Return ID kparameter pvname."""
        config = self._value
        if config:
            kparm = config['pvnames']['kparameter']
            return kparm
        else:
            raise ValueError('Configuration not defined!')

    @property
    def ch_pvnames(self):
        """Return CH corrector power supply pvnames."""
        return self._get_corr_pvnames('ch1', 'ch2')

    @property
    def cv_pvnames(self):
        """Return CV corrector power supply pvnames."""
        return self._get_corr_pvnames('cv1', 'cv2')

    @property
    def qs_pvnames(self):
        """Return QS corrector power supply pvnames."""
        return self._get_corr_pvnames('qs1', 'qs2')

    @property
    def polarizations(self):
        """Return list of light polarizations in the IDFF config."""
        if self._value:
            return list(self._value['polarizations'].keys())
        else:
            raise ValueError('Configuration not defined!')

    @property
    def value(self):
        """Get configuration."""
        return _dcopy(self._value)

    @value.setter
    def value(self, value):
        """Set configuration."""
        self._set_value(value)

    def calculate_setpoints(
            self, polarization, pparameter_value, kparameter_value):
        """Return correctors setpoints for a particular ID config.

        'pparameter' is the ID phase which defines polarization.
        'kparameter' can be a gap or phase value, depending on the ID.
        """
        if self._value:
            setpoints = dict()
            idff = self._value['polarizations'][polarization]
            if polarization == 'none':
                params = idff['pparameter']
                param_value = pparameter_value
            else:
                params = idff['kparameter']
                param_value = kparameter_value
            setpoints = dict()
            for corrlabel, table in idff.items():
                if corrlabel not in ('pparameter', 'kparameter'):
                    # linear interpolation
                    setpoint = _np.interp(param_value, params, table)
                    corr_pvname = self._value['pvnames'][corrlabel]
                    setpoints[corr_pvname] = setpoint
            return setpoints
        else:
            raise ValueError('Configuration not defined!')

    @staticmethod
    def create_template_config(idname):
        """Create a template configuration for a given ID type."""
        idff = _IDSearch.conv_idname_2_idff(idname)

        # description
        description = 'Config created with IDFFConfig.create_template_config'

        # build value for pvnames key
        pvnames = dict()
        for key, corr_pvname in idff.items():
            if key != 'polarizations':
                pvnames[key] = corr_pvname

        # build value for polarizations key
        ctable = dict()
        for pvname in pvnames:
            ctable[pvname] = [0, 0]
        ptable = dict(ctable)
        ptable['pparameter'] = [0, 0]
        ptable['kparameter'] = 0
        ktable = dict(ctable)
        ktable['kparameter'] = [0, 0]

        pol2pparam = _IDSearch.conv_idname_2_polarization_pparameter
        polarizations = dict()
        for pol in idff['polarizations']:
            if pol == _IDSearch.POL_UNDEF_STR:
                continue
            if pol == _IDSearch.POL_NONE_STR:
                polarizations[pol] = dict(ptable)
            else:
                ktable['pparameter'] = pol2pparam(idname, pol)
                polarizations[pol] = dict(ktable)

        template_config = dict(
            description=description, pvnames=pvnames,
            polarizations=polarizations)
        return template_config

    def __str__(self):
        """."""
        stg = ''
        stg += f'name: {self.name}'
        value = self.value
        if value is None:
            return stg
        stg += f'\ndescription: {value["description"]}'

        stg += '\n--- pvnames ---'
        pvnames = ''.join(
            [f'\n{key}: {value}' for key, value in value['pvnames'].items()])
        stg += pvnames

        for pol, table in self.value['polarizations'].items():
            stg += f'\n--- {pol} ---'
            for key, value in table.items():
                if isinstance(value, (int, float)):
                    nrpts = 1
                    str_ = f'{value}'
                else:
                    nrpts = len(value)
                    str_ = f'{min(value)}..{max(value)} [{nrpts}]'
                stg += f'\n{key} : ' + str_
        return stg

    def load(self, discarded=False):
        """."""
        super().load(discarded=discarded)
        self._find_idname()
        self._calc_polariz_defs()

    def get_polarization_state(self, pparameter, kparameter):
        """Return polarization state based on ID parameteres."""
        idname = self._idname
        pol_idx = _IDSearch.conv_idname_2_polarization_state(
            idname, pparameter, kparameter)
        pol_str = _IDSearch.conv_idname_2_polarizations_sts(idname)[pol_idx]
        return pol_str

    def check_valid_value(self, value):
        """Check consistency of SI_IDFF configuration."""
        configs = value['polarizations']
        pvnames = {
            key: value for key, value in value['pvnames'].items()
            if key not in ('pparameter', 'kparameter')}
        corrlabels = set(pvnames.keys())

        # check pvnames in configs
        pvsconfig = set(pvnames.values())
        getch = _IDSearch.conv_idname_2_idff_chnames
        getcv = _IDSearch.conv_idname_2_idff_cvnames
        getqs = _IDSearch.conv_idname_2_idff_qsnames
        chnames = [corr + ':Current-SP' for corr in getch(self.idname)]
        cvnames = [corr + ':Current-SP' for corr in getcv(self.idname)]
        qsnames = [corr + ':Current-SP' for corr in getqs(self.idname)]
        pvsidsearch = set(chnames + cvnames + qsnames)
        symm_diff = pvsconfig ^ pvsidsearch

        if symm_diff:
            raise ValueError('List of pvnames in config is not consistent')

        # check polarizations in configs
        pconfig = set(configs.keys()) - set((_IDSearch.POL_NONE_STR, ))
        pidsearch = set(_IDSearch.conv_idname_2_polarizations(self.idname))
        symm_diff = pconfig ^ pidsearch

        if symm_diff:
            raise ValueError(
                'List of polarizations in config is not consistent')

        # check polarization tables consistency
        for polarization, table in configs.items():
            corrtable = {
                key: value for key, value in table.items()
                if key not in ('pparameter', 'kparameter')}

            # check 'pparameter'
            if 'pparameter' not in table:
                raise ValueError(
                    'Missing pparameter in polarization configuration.')

            # check 'kparameter'
            if 'kparameter' not in table:
                raise ValueError(
                    'Missing kparameter in polarization configuration.')

            # check corr label list
            corrlabels_config = set(corrtable.keys())
            symm_diff = corrlabels ^ corrlabels_config
            if symm_diff:
                raise ValueError(
                    'List of corrlabels in config is not consistent')

            # check nrpts in tables
            param = 'pparameter' if polarization == 'none' else 'kparameter'
            nrpts_corrtables = {len(table) for table in corrtable.values()}
            nrpts_kparameter = set([len(table[param]), ])
            symm_diff = nrpts_corrtables ^ nrpts_kparameter

            if symm_diff:
                raise ValueError(
                    'Corrector tables and kparameter list in config '
                    'are not consistent')
        return True

    def _get_corr_pvnames(self, cname1, cname2):
        """Return corrector power supply pvnames."""
        if self._value:
            pvnames = self._value['pvnames']
            corr1, corr2 = pvnames.get(cname1), pvnames.get(cname2)
            return corr1, corr2
        else:
            raise ValueError('Configuration not defined!')

    def _set_value(self, value):
        super()._set_value(value)
        self._find_idname()
        self._calc_polariz_defs()

    def _calc_polariz_defs(self):
        """."""
        # fill polarization data struct
        data = self._value['polarizations']
        poldefs = dict()
        for pol, tab in data.items():
            if pol != _IDSearch.POL_NONE_STR:
                poldefs[pol] = tab['pparameter']
            else:
                poldefs[pol] = tab['kparameter']
        self._polarization_definitions = poldefs

    def _find_idname(self):
        """."""
        # find associated idname
        self._idname = None
        pvnames = self._value['pvnames']
        kparameter, pparameter = pvnames['kparameter'], pvnames['pparameter']
        for idname in _IDSearch.get_idnames():
            kparam_propty = _IDSearch.conv_idname_2_kparameter_propty(idname)
            pparam_propty = _IDSearch.conv_idname_2_pparameter_propty(idname)
            if None in (kparam_propty, pparam_propty):
                continue
            kparam = idname + ':' + kparam_propty
            pparam = idname + ':' + pparam_propty

            if kparam == kparameter and pparam == pparameter:
                self._idname = idname
                break
        if self._idname is None:
            # could not find idname
            raise ValueError(
                'kparameter and pparameter in config are not '
                'associated with an idname!')



#!/usr/bin/env python-sirius

"""Unittest module for util.py."""

from io import StringIO
import logging
from unittest import mock, TestCase
import numpy as np

import siriuspy.util as util


PUB_INTERFACE = (
    'conv_splims_labels',
    'get_last_commit_hash',
    'get_timestamp',
    'read_text_data',
    'print_ioc_banner',
    'configure_log_file',
    'beam_rigidity',
    'check_pv_online',
    'get_strength_label',
    'get_strength_units',
    'update_bit',
    'get_bit',
    'mode',
    'check_public_interface_namespace',
    'ClassProperty',
    )


class TestUtil(TestCase):
    """Test util."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(util, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_conv_splims_labels(self):
        """Test conv_splims_labels."""
        self.assertEqual(util.conv_splims_labels('HOPR'), 'hilim')
        self.assertEqual(util.conv_splims_labels('lolo'), 'LOLO')
        self.assertRaises(KeyError, util.conv_splims_labels, label='dummmy')

    def test_get_last_commit_hash(self):
        """Test get_signal_names."""
        lst_comm = util.get_last_commit_hash()
        self.assertIsInstance(lst_comm, str)

    def test_get_timestamp(self):
        """Test get_timestamp."""
        stime = util.get_timestamp()
        self.assertEqual(len(stime), 23)
        date, minute, second = stime.split(':')
        year, month, day, hour = date.split('-')
        second, msecond = second.split('.')
        self.assertTrue(year.isnumeric())
        self.assertTrue(month.isnumeric())
        self.assertTrue(day.isnumeric())
        self.assertTrue(hour.isnumeric())
        self.assertTrue(minute.isnumeric())
        self.assertTrue(second.isnumeric())
        self.assertTrue(msecond.isnumeric())
        self.assertGreater(int(year), 2016)
        self.assertLess(int(month), 13)
        self.assertGreater(int(month), 0)
        self.assertLess(int(day), 32)
        self.assertGreater(int(day), 0)
        self.assertLess(int(hour), 24)
        self.assertGreater(int(hour), -1)
        self.assertLess(int(minute), 60)
        self.assertGreater(int(minute), -1)
        self.assertLess(int(second), 60)
        self.assertGreater(int(second), -1)
        self.assertLess(int(msecond), 1000)
        self.assertGreater(int(msecond), -1)

    def test_read_text_data(self):
        """Test read_text_data."""
        text = '\n'
        text += '# comment line 1\n'
        text += '# comment line 2 #\n'
        text += '# comment line 3 ## - test text !@#$&!&\n'
        text += '\n'
        text += '# comment line 4 after empty line\n'
        text += '\n'
        text += '\n'
        text += ' # comment line 5 starting wih space\n'
        text += '     # comment line 6 starting wih multiple spaces\n'
        text += ' # [parameter1] v1\n'  # comment line parameter
        text += ' #[parameter2] v2  \n'  # comment line with parameter
        text += '# [parameter3] v3 v4 v5  \n'  # comment line with parameter
        text += 'datum1 datum2 datum3\n'  # line with data
        text += 'datum4 datum5 datum6\n'  # line with data
        text += '\n'  # empty line between data
        text += '# comment line 5 after empty line\n'  # comment between data
        text += 'datum7\n'  # line with data
        _, params = util.read_text_data(text)
        self.assertEqual(len(params), 3)
        self.assertEqual(len(params['parameter1']), 1)
        self.assertEqual(params['parameter1'][0], 'v1')
        self.assertEqual(len(params['parameter2']), 1)
        self.assertEqual(params['parameter2'][0], 'v2')
        self.assertEqual(len(params['parameter3']), 3)
        self.assertEqual(params['parameter3'][0], 'v3')
        self.assertEqual(params['parameter3'][1], 'v4')
        self.assertEqual(params['parameter3'][2], 'v5')

    def test_print_ioc_banner(self):
        """Test print_ioc_banner."""
        dbase = {}
        dbase.update(
            {'PV01': None, 'PV02': None, 'PV03': None, 'PV04': None, })
        dbase.update(
            {'PV05': None, 'PV06': None, 'PV07': None, 'PV08': None, })
        fil = StringIO()
        logging.root.handlers = []
        util.configure_log_file(stream=fil)
        util.print_ioc_banner(
            'test-ioc', dbase, 'Test-ioc for util module unittest',
            '1.0.0', 'PREFIX')
        text = fil.getvalue()
        self.assertEqual(len(text.splitlines()), 17)
        dbase.update(
            {'PV09': None, 'PV10': None, 'PV11': None, 'PV12': None, })
        dbase.update(
            {'PV13': None, 'PV14': None, 'PV15': None, 'PV16': None, })
        util.print_ioc_banner(
            'test-ioc', dbase, 'Test-ioc for util module unittest',
            '1.0.0', 'PREFIX')
        text = fil.getvalue()
        self.assertEqual(len(text.splitlines()), 25+17)

    def test_configure_log_file(self):
        """Test configure_log_file."""
        fil = StringIO()
        logging.root.handlers = []
        util.configure_log_file(stream=fil)
        logging.info('test')
        logging.debug('test')
        text = fil.getvalue()
        self.assertEqual(len(text.splitlines()), 1)

        fil = StringIO()
        logging.root.handlers = []
        util.configure_log_file(stream=fil, debug=True)
        logging.info('test')
        logging.debug('test')
        text = fil.getvalue()
        self.assertEqual(len(text.splitlines()), 2)

    def test_beam_rigidity(self):
        """Test beam_rigidity."""
        en0 = 510998.92760316096/1e9  # electron rest energy [GeV]
        tol = 1e-12
        # float arg
        ret, beta, gamma = util.beam_rigidity(3.0)
        self.assertIsInstance(ret, float)
        self.assertAlmostEqual(ret, 10.00692271077752, delta=tol)
        self.assertAlmostEqual(beta, 0.9999999854933386, delta=tol)
        self.assertAlmostEqual(gamma, 5870.853550721619, delta=tol)
        ret, *_ = util.beam_rigidity(energy=1.001*en0)
        self.assertIsInstance(ret, float)
        self.assertAlmostEqual(ret, 7.624534235479652e-05, delta=tol)
        ret, *_ = util.beam_rigidity(0.0)
        self.assertEqual(ret, 0.0)
        ret, *_ = util.beam_rigidity(0.999*en0)
        self.assertEqual(ret, 0.0)
        # list arg
        ret, *_ = util.beam_rigidity([0.5, 3.0])
        self.assertIsInstance(ret, np.ndarray)
        self.assertAlmostEqual(ret[0], 1.6678196049882876, delta=tol)
        self.assertAlmostEqual(ret[1], 10.00692271077752, delta=tol)
        # tuple arg
        ret, *_ = util.beam_rigidity((0.5, 3.0))
        self.assertIsInstance(ret, np.ndarray)
        self.assertAlmostEqual(ret[0], 1.6678196049882876, delta=tol)
        self.assertAlmostEqual(ret[1], 10.00692271077752, delta=tol)
        # numpy.ndarray arg
        ret, *_ = util.beam_rigidity(np.array([0.5, 3.0]))
        self.assertIsInstance(ret, np.ndarray)
        self.assertAlmostEqual(ret[0], 1.6678196049882876, delta=tol)
        self.assertAlmostEqual(ret[1], 10.00692271077752, delta=tol)
        ret, *_ = util.beam_rigidity(energy=np.array([1.001*en0, 3.0]))
        self.assertIsInstance(ret, np.ndarray)
        self.assertAlmostEqual(ret[0], 7.624534235479652e-05, delta=tol)

    def test_check_pv_online(self):
        """Test check_pv_online."""
        with mock.patch('siriuspy.util._epics.PV') as mock_conn:
            mock_conn.return_value.wait_for_connection.return_value = True
            status = util.check_pv_online("FakePV")
        self.assertEqual(status, True)
        mock_conn.assert_called()
        with mock.patch('siriuspy.util._epics.PV') as mock_conn:
            mock_conn.return_value.wait_for_connection.return_value = False
            status = util.check_pv_online("FakePV")
        self.assertEqual(status, False)
        mock_conn.assert_called()

    def test_get_strength_label(self):
        """Test get_strength_label."""
        self.assertEqual("Energy", util.get_strength_label("dipole"))
        self.assertEqual("KL", util.get_strength_label("quadrupole"))
        self.assertEqual("KL", util.get_strength_label("quadrupole-skew"))
        self.assertEqual("SL", util.get_strength_label("sextupole"))
        self.assertEqual(
            "Kick", util.get_strength_label("corrector-horizontal"))
        self.assertEqual("Kick", util.get_strength_label("corrector-vertical"))
        self.assertRaises(ValueError, util.get_strength_label, magfunc='')

    def test_get_strength_units(self):
        """Test get_strength_units."""
        self.assertEqual("GeV", util.get_strength_units("dipole"))
        self.assertEqual("1/m", util.get_strength_units("quadrupole"))
        self.assertEqual("1/m", util.get_strength_units("quadrupole-skew"))
        self.assertEqual("1/m^2", util.get_strength_units("sextupole"))
        self.assertEqual(
            "urad",
            util.get_strength_units("corrector-horizontal", None, False))
        self.assertEqual(
            "urad", util.get_strength_units("corrector-vertical", None, False))
        self.assertEqual(
            "mrad",
            util.get_strength_units("corrector-horizontal", None, True))
        self.assertEqual(
            "mrad", util.get_strength_units("corrector-vertical", None, True))
        self.assertRaises(ValueError, util.get_strength_units, magfunc='')

    def test_update_bit(self):
        """Test update_bit."""
        with self.assertRaises(TypeError):
            util.update_bit(v=0.1, bit_pos=0, bit_val=0)
        with self.assertRaises(TypeError):
            util.update_bit(v=0x2, bit_pos=1.0, bit_val=2)
        with self.assertRaises(ValueError):
            util.update_bit(v=-5, bit_pos=1, bit_val=0)
        with self.assertRaises(ValueError):
            util.update_bit(v=5, bit_pos=-1, bit_val=0)
        self.assertEqual(0xff, util.update_bit(v=0xdf, bit_pos=5, bit_val=1))
        self.assertEqual(0x00, util.update_bit(v=0x10, bit_pos=4, bit_val=0))
        self.assertEqual(0x00, util.update_bit(v=0x10, bit_pos=4, bit_val=[]))
        self.assertEqual(0x10, util.update_bit(v=0x10, bit_pos=4, bit_val=45))

    def test_get_bit(self):
        """Test get_bit."""
        with self.assertRaises(TypeError):
            util.get_bit(v=0.1, bit_pos=0)
        with self.assertRaises(TypeError):
            util.get_bit(v=0x2, bit_pos=1.0)
        with self.assertRaises(ValueError):
            util.get_bit(v=-5, bit_pos=1)
        with self.assertRaises(ValueError):
            util.get_bit(v=5, bit_pos=-1)
        self.assertEqual(0, util.get_bit(v=0xdf, bit_pos=5))
        self.assertEqual(1, util.get_bit(v=0x10, bit_pos=4))

    def test_mode(self):
        """Test test_mode."""
        self.assertEqual((3, 2), util.mode([1, 2, 3, 4, 5, 6, 3]))

    def test_check_public_interface_namespace(self):
        """Test check_public_interface_namespace."""
        class NameSpace:
            """."""

            A = None
            B = None

        public_interface = ('A', 'B', )
        valid = util.check_public_interface_namespace(
            NameSpace, public_interface)
        self.assertTrue(valid)
        NameSpace.C = None
        valid = util.check_public_interface_namespace(
            NameSpace, public_interface, print_flag=False)
        self.assertFalse(valid)
        public_interface = ('A', 'B', 'missing_symbol')
        valid = util.check_public_interface_namespace(
            NameSpace, public_interface, print_flag=False)
        self.assertFalse(valid)



"""Translate `ps_state` bits to values exposed in EPICS."""

from ..csdev import ETypes as _et
from ..csdev import Const as _c


class PSCStatus:
    """Power Supply Controller Status."""

    OPMODE = _c.OpMode
    PWRSTATE = _c.PwrStateSel
    STATES = _c.States
    CYCLETYPE = _c.CycleType

    _mask_state = 0b0000000000001111
    _mask_oloop = 0b0000000000010000
    _mask_intfc = 0b0000000001100000
    _mask_activ = 0b0000000010000000
    _mask_model = 0b0001111100000000
    _mask_unlck = 0b0010000000000000
    _mask_rsrvd = 0b1100000000000000
    _mask_stats = 0b1111111111111111

    _psc2ioc_state = {
        _c.States.Off: _c.OpMode.SlowRef,
        _c.States.Interlock: _c.OpMode.SlowRef,
        _c.States.Initializing: _c.OpMode.SlowRef,
        _c.States.SlowRef: _c.OpMode.SlowRef,
        _c.States.SlowRefSync: _c.OpMode.SlowRefSync,
        _c.States.Cycle: _c.OpMode.Cycle,
        _c.States.RmpWfm: _c.OpMode.RmpWfm,
        _c.States.MigWfm: _c.OpMode.MigWfm,
        _c.States.FastRef: _c.OpMode.FastRef,
    }

    _ioc2psc_state = {
        # TODO: controller firmware still defines only a subset of opmodes
        _c.OpMode.SlowRef: _c.States.SlowRef,
        _c.OpMode.SlowRefSync: _c.States.SlowRefSync,
        _c.OpMode.Cycle: _c.States.Cycle,
        _c.OpMode.RmpWfm: _c.States.SlowRef,
        _c.OpMode.MigWfm: _c.States.SlowRef,
        _c.OpMode.FastRef: _c.States.SlowRef,
    }

    def __init__(self, ps_status=0):
        """Constructor."""
        self._ps_status = ps_status

    # --- public interface ---

    @property
    def ps_status(self):
        """Return ps-controller ps_status."""
        return self._ps_status

    @ps_status.setter
    def ps_status(self, value):
        """Set ps-controller ps_status."""
        self._ps_status = value & PSCStatus._mask_stats

    @property
    def state(self):
        """Return ps-controller state."""
        return (self._ps_status & PSCStatus._mask_state) >> 0

    @state.setter
    def state(self, value):
        """Set ps-controller state."""
        self._ps_status = self._ps_status & ~PSCStatus._mask_state
        self._ps_status += (value & 0b1111) << 0

    @property
    def open_loop(self):
        """Return ps-controller open_loop."""
        return (self._ps_status & PSCStatus._mask_oloop) >> 4

    @open_loop.setter
    def open_loop(self, value):
        """Set ps-controller open_loop."""
        self._ps_status = self._ps_status & ~PSCStatus._mask_oloop
        self._ps_status += (value & 0b1) << 4

    @property
    def interface(self):
        """Return ps-controller interface."""
        return (self._ps_status & PSCStatus._mask_intfc) >> 5

    @interface.setter
    def interface(self, value):
        """Set ps-controller interface."""
        self._ps_status = self._ps_status & ~PSCStatus._mask_intfc
        self._ps_status += (value & 0b11) << 5

    @property
    def active(self):
        """Return ps-controller active."""
        return (self._ps_status & PSCStatus._mask_activ) >> 7

    @active.setter
    def active(self, value):
        """Set ps-controller active."""
        self._ps_status = self._ps_status & ~PSCStatus._mask_activ
        self._ps_status += (value & 0b1) << 7

    @property
    def model(self):
        """Return ps-controller model."""
        return (self._ps_status & PSCStatus._mask_model) >> 8

    @model.setter
    def model(self, value):
        """Set ps-controller interface."""
        self._ps_status = self._ps_status & ~PSCStatus._mask_model
        self._ps_status += (value & 0b11111) << 8

    @property
    def unlocked(self):
        """Return ps-controller unlocked."""
        return (self._ps_status & PSCStatus._mask_unlck) >> 13

    @unlocked.setter
    def unlocked(self, value):
        """Set ps-controller unlocked."""
        self._ps_status = self._ps_status & ~PSCStatus._mask_unlck
        self._ps_status += (value & 0b1) << 13

    @property
    def reserved(self):
        """Return ps-controller reserved."""
        return (self._ps_status & PSCStatus._mask_rsrvd) >> 14

    @reserved.setter
    def reserved(self, value):
        """Set ps-controller reserved."""
        self._ps_status = self._ps_status & ~PSCStatus._mask_rsrvd
        self._ps_status += (value & 0b11) << 14

    @property
    def ioc_pwrstate(self):
        """Return ioc-controller power state."""
        state = self.state
        if state in (_c.States.Off,
                     _c.States.Interlock):
            pwrstate = _c.PwrStateSel.Off
        else:
            pwrstate = _c.PwrStateSel.On
        return pwrstate

    @ioc_pwrstate.setter
    def ioc_pwrstate(self, value):
        """Set ps_status with a given ioc-controller power state."""
        if not 0 <= value < len(_et.PWRSTATE_SEL):
            raise ValueError('Invalid pwrstate value!')
        # TurnOn sets Opmode to SlowRef by default.
        state = _c.States.Off if value == _c.PwrStateSel.Off else \
            _c.States.SlowRef
        self.state = state

    @property
    def ioc_opmode(self):
        """Return ioc-controller opmode."""
        state = self.state
        opmode = PSCStatus._psc2ioc_state[state]
        return opmode

    @ioc_opmode.setter
    def ioc_opmode(self, value):
        """Set ps_status with a given ioc-controller opmode."""
        if not 0 <= value < len(_et.OPMODES):
            raise ValueError('Invalid opmode value!')
        state = PSCStatus._ioc2psc_state[value]
        self.state = state



"""PosAng utilities."""
import os as _os


class HandleConfigNameFile:
    """Class to handle config_name file."""

    def __init__(self, tl, corrs_type):
        """Init."""
        if not (tl in ['TB', 'TS']):
            raise ValueError("'acc' must be 'TB' or 'TS' instance.")
        if not (corrs_type in ['ch-sept', 'sept-sept']):
            raise ValueError(
                "'opticsparam' must be 'ch-sept' or 'sept-sept' instance.")
        self.tl = tl.lower()
        self.corrs_type = corrs_type
        self.fpath = '/home/sirius/iocs-log/' + self.tl + '-ap-posang/'
        self.fname = self.fpath + self.tl + '-posang.txt'

    def get_config_name(self):
        """Get config name from file."""
        if _os.path.isfile(self.fname):
            with open(self.fname, 'r') as fil:
                config_name = fil.read().strip('\n')
        else:
            _os.makedirs(self.fpath, exist_ok=True)
            config_name = get_default_config_name(self.tl, self.corrs_type)
            with open(self.fname, 'w+') as fil:
                fil.write(config_name)
        return config_name

    def set_config_name(self, config_name):
        """Set config name in file."""
        if not isinstance(config_name, str):
            raise TypeError("'config_name' must be a string.")
        with open(self.fname, 'w+') as fil:
            fil.write(config_name)


def get_default_config_name(tl, corrs_type):
    """Get default config name."""
    if tl.upper() == 'TB':
        return 'Default_CHSept'
    else:
        if corrs_type == 'sept-sept':
            return 'TS.V04.01-M1.SeptSept.4CVs'
        else:
            return 'TS.V04.01-M1.CHSept.4CVs'



"""Main module of AS-AP-TuneCorr IOC."""

import numpy as _np
from epics import PV as _PV

from ..envars import VACA_PREFIX as _vaca_prefix
from ..namesys import SiriusPVName as _SiriusPVName

from .csdev import Const as _Const
from .base import BaseApp as _BaseApp


class TuneCorrApp(_BaseApp):
    """Main application for handling tune correction."""

    _optics_param = 'tune'

    def __init__(self, acc):
        """Class constructor."""
        super().__init__(acc)

        # consts
        self._delta_tunex = 0.0
        self._delta_tuney = 0.0

        self._set_new_refkl_cmd_count = 0

        if self._acc == 'SI':
            self._meas_config_dkl_qf = 0.020
            self._meas_config_dkl_qd = 0.020

        # Connect to Quadrupoles Families
        self._psfam_refkl = {fam: 0 for fam in self._psfams}
        self._lastcalc_deltakl = {fam: 0 for fam in self._psfams}
        for fam in self._psfams:
            pvname = _SiriusPVName(self._acc+'-Fam:PS-'+fam+':KL-RB')
            pvname = pvname.substitute(prefix=_vaca_prefix)
            self._psfam_intstr_rb_pvs[fam] = _PV(
                pvname,
                callback=[self._callback_init_refkl,
                          self._callback_estimate_deltatune],
                connection_timeout=0.05)

        self.map_pv2write.update({
            'DeltaTuneX-SP': self.set_dtune_x,
            'DeltaTuneY-SP': self.set_dtune_y,
            'SetNewRefKL-Cmd': self.cmd_set_newref,
            'MeasConfigDeltaKLFamQF-SP': self.set_meas_config_dkl_qf,
            'MeasConfigDeltaKLFamQD-SP': self.set_meas_config_dkl_qd,
        })

    def update_corrparams_pvs(self):
        """Set initial correction parameters PVs values."""
        self.run_callbacks('RespMat-Mon', self._nominal_matrix)
        self.run_callbacks('NominalKL-Mon', self._psfam_nom_intstr)

    # ------ handle pv write methods -------

    def set_dtune_x(self, value):
        """Set DeltaTuneX."""
        self._delta_tunex = value
        self.run_callbacks('DeltaTuneX-RB', value)
        self._calc_intstrength()
        return True

    def set_dtune_y(self, value):
        """Set DeltaTuneY."""
        self._delta_tuney = value
        self.run_callbacks('DeltaTuneY-RB', value)
        self._calc_intstrength()
        return True

    def cmd_set_newref(self, value):
        """SetNewRefKL command."""
        if self._update_ref():
            self._set_new_refkl_cmd_count += 1
            self.run_callbacks(
                'SetNewRefKL-Cmd', self._set_new_refkl_cmd_count)
        return False

    def set_meas_config_dkl_qf(self, value):
        """Set MeasConfigDeltaKLFamQF."""
        if value == self._meas_config_dkl_qf:
            return False
        self._meas_config_dkl_qf = value
        self.run_callbacks('MeasConfigDeltaKLFamQF-RB', value)
        return True

    def set_meas_config_dkl_qd(self, value):
        """Set MeasConfigDeltaKLFamQD."""
        if value == self._meas_config_dkl_qd:
            return False
        self._meas_config_dkl_qd = value
        self.run_callbacks('MeasConfigDeltaKLFamQD-RB', value)
        return True

    # ---------- auxiliar methods ----------

    def _handle_corrparams_2_read(self, params):
        """Edit correction params."""
        nom_matrix = [item for sublist in params['matrix'] for item in sublist]
        nom_kl = params['nominal KLs']
        nom_deltakl = [0.0, 0.0]
        return nom_matrix, nom_kl, nom_deltakl

    def _handle_corrparams_2_save(self):
        matrix = _np.array(self._nominal_matrix)
        matrix = _np.reshape(matrix, [2, len(self._psfams)])

        value = {'matrix': matrix,
                 'nominal KLs': self._psfam_nom_intstr}
        return value

    def _calc_intstrength(self):
        method = 0 \
            if self._corr_method == _Const.CorrMeth.Proportional \
            else 1
        grouping = '2knobs' \
            if self._corr_group == _Const.CorrGroup.TwoKnobs \
            else 'svd'
        lastcalc_deltakl = self._opticscorr.calculate_delta_intstrengths(
            method=method, grouping=grouping,
            delta_opticsparam=[self._delta_tunex, self._delta_tuney])

        self.run_callbacks('Log-Mon', 'Calculated KL values.')

        for fam_idx, fam in enumerate(self._psfams):
            self._lastcalc_deltakl[fam] = lastcalc_deltakl[fam_idx]
            self.run_callbacks(
                'DeltaKL'+fam+'-Mon', self._lastcalc_deltakl[fam])

    def _apply_corr(self):
        if self._is_status_ok():
            kls = {fam: self._psfam_refkl[fam]+self._lastcalc_deltakl[fam]
                   for fam in self._psfams}
            self._apply_intstrength(kls)
            self.run_callbacks('Log-Mon', 'Applied correction.')

            if self._sync_corr == _Const.SyncCorr.On:
                self._event_exttrig_cmd.put(0)
                self.run_callbacks('Log-Mon', 'Generated trigger.')
            return True

        self.run_callbacks('Log-Mon', 'ERR: ApplyDelta-Cmd failed.')
        return False

    def _get_optics_param(self):
        """Return optics parameter."""
        return self._get_tunes()

    def _get_delta_intstrength(self, fam):
        """Get delta to apply in each family."""
        if 'QF' in fam:
            deltakl = self._meas_config_dkl_qf
        else:
            deltakl = self._meas_config_dkl_qd
        fam_idx = self._psfams.index(fam)
        nelm = self._psfam_nelm[fam_idx]
        return deltakl/nelm

    def _update_ref(self):
        if (self._status & 0x1) == 0:  # Check connection
            # update references
            for fam in self._psfams:
                value = self._psfam_intstr_rb_pvs[fam].get()
                if value is None:
                    self.run_callbacks(
                        'Log-Mon',
                        'ERR: Received a None value from {}.'.format(fam))
                    return False
                self._psfam_refkl[fam] = value
                self.run_callbacks(
                    'RefKL' + fam + '-Mon', self._psfam_refkl[fam])

                self._lastcalc_deltakl[fam] = 0
                self.run_callbacks('DeltaKL' + fam + '-Mon', 0)

            # the deltas from new kl references are zero
            self._delta_tunex = 0
            self._delta_tuney = 0
            self.run_callbacks('DeltaTuneX-SP', self._delta_tunex)
            self.run_callbacks('DeltaTuneX-RB', self._delta_tunex)
            self.run_callbacks('DeltaTuneY-SP', self._delta_tuney)
            self.run_callbacks('DeltaTuneY-RB', self._delta_tuney)

            self._estimate_current_deltatune()

            self.run_callbacks('Log-Mon', 'Updated KL references.')
            return True

        self.run_callbacks(
            'Log-Mon', 'ERR: Some magnet family is disconnected.')
        return False

    def _estimate_current_deltatune(self):
        psfam_deltakl = len(self._psfams)*[0]
        for fam_idx, fam in enumerate(self._psfams):
            psfam_deltakl[fam_idx] = \
                self._psfam_intstr_rb[fam] - self._psfam_refkl[fam]
        self._optprm_est = self._opticscorr.calculate_opticsparam(
            psfam_deltakl)
        self.run_callbacks('DeltaTuneX-Mon', self._optprm_est[0])
        self.run_callbacks('DeltaTuneY-Mon', self._optprm_est[1])

    # ---------- callbacks ----------

    def _callback_init_refkl(self, pvname, value, cb_info, **kws):
        """Initialize RefKL-Mon pvs and remove this callback."""
        # Get reference
        if value is None:
            return
        fam = _SiriusPVName(pvname).dev
        self._psfam_refkl[fam] = value
        self.run_callbacks('RefKL'+fam+'-Mon', self._psfam_refkl[fam])

        # Remove callback
        cb_info[1].remove_callback(cb_info[0])

    def _callback_estimate_deltatune(self, pvname, value, **kws):
        if value is None:
            return
        fam = _SiriusPVName(pvname).dev
        self._psfam_intstr_rb[fam] = value
        self._estimate_current_deltatune()



"""ID feedforward database."""

import os as _os

from .. import csdev as _csdev
from ..namesys import SiriusPVName as _PVName
from ..search import IDSearch as _IDSearch


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    OPEN_CLOSED = ('Open', 'Closed')
    STS_LBLS_CORR = (
        'Connected', 'PwrStateOn', 'OpModeConfigured')


_et = ETypes


# --- Const class ---

class IDFFConst(_csdev.Const):
    """ID Feedforward Const class."""

    LoopState = _csdev.Const.register('LoopState', _et.OPEN_CLOSED)
    StsLblsCorr = _csdev.Const.register(
        'StsLblsCorr', _et.STS_LBLS_CORR)

    DEFAULT_CORR_STATUS = 0b11111111
    DEFAULT_LOOP_FREQ_MIN = 0.001  # [Hz]
    DEFAULT_LOOP_FREQ_MAX = 100  # [Hz]
    DEFAULT_LOOP_FREQ = 10  # [Hz]
    DEFAULT_LOOP_STATE = LoopState.Open
    DEFAULT_CONTROL_QS = _csdev.Const.DsblEnbl.Enbl
    DEFAULT_CORR_PREC = 4

    def __init__(self, idname):
        """Init."""
        self.idname = _PVName(idname)
        self.idffname = 'SI-' + self.idname.sub + ':AP-IDFF'
        ioc_fol = _os.path.join(
            '/home', 'sirius', 'iocs-log', 'si-ap-idff', 'data')
        fname = '_'.join([self.idname.sec, self.idname.sub, self.idname.dev])
        fname = fname.lower()
        self.autosave_fname = _os.path.join(ioc_fol, fname+'.txt')
        qsnames = _IDSearch.conv_idname_2_idff_qsnames(idname)
        self.has_qscorrs = True if qsnames else False

    def get_propty_database(self):
        """Return property database."""
        dbase = {
            'Version-Cte': {'type': 'str', 'value': 'UNDEF'},
            'Log-Mon': {'type': 'string', 'value': 'Starting...'},
            'LoopState-Sel': {
                'type': 'enum', 'enums': _et.OPEN_CLOSED,
                'value': self.DEFAULT_LOOP_STATE},
            'LoopState-Sts': {
                'type': 'enum', 'enums': _et.OPEN_CLOSED,
                'value': self.DEFAULT_LOOP_STATE},
            'LoopFreq-SP': {
                'type': 'float', 'value': self.DEFAULT_LOOP_FREQ,
                'unit': 'Hz', 'prec': 3,
                'lolim': self.DEFAULT_LOOP_FREQ_MIN,
                'hilim': self.DEFAULT_LOOP_FREQ_MAX},
            'LoopFreq-RB': {
                'type': 'float', 'value': self.DEFAULT_LOOP_FREQ,
                'unit': 'Hz', 'prec': 3,
                'lolim': self.DEFAULT_LOOP_FREQ_MIN,
                'hilim': self.DEFAULT_LOOP_FREQ_MAX},
            'Polarization-Mon': {'type': 'string', 'value': 'none'},
            'ConfigName-SP': {'type': 'string', 'value': ''},
            'ConfigName-RB': {'type': 'string', 'value': ''},
            'CorrConfig-Cmd': {'type': 'int', 'value': 0},
            'CorrStatus-Mon': {
                'type': 'int', 'value': self.DEFAULT_CORR_STATUS},
            'CorrStatusLabels-Cte': {
                'type': 'string', 'count': len(self.StsLblsCorr._fields),
                'value': self.StsLblsCorr._fields},
            'CorrCH1Current-Mon': {
                'type': 'float', 'value': 0,
                'unit': 'A', 'prec': self.DEFAULT_CORR_PREC},
            'CorrCH2Current-Mon': {
                'type': 'float', 'value': 0,
                'unit': 'A', 'prec': self.DEFAULT_CORR_PREC},
            'CorrCV1Current-Mon': {
                'type': 'float', 'value': 0,
                'unit': 'A', 'prec': self.DEFAULT_CORR_PREC},
            'CorrCV2Current-Mon': {
                'type': 'float', 'value': 0,
                'unit': 'A', 'prec': self.DEFAULT_CORR_PREC},
        }
        if self.has_qscorrs:
            dbase.update({
                'ControlQS-Sel': {
                    'type': 'enum', 'enums': _et.DSBL_ENBL,
                    'value': self.DEFAULT_CONTROL_QS,
                    'unit': 'If QS are included in loop'},
                'ControlQS-Sts': {
                    'type': 'enum', 'enums': _et.DSBL_ENBL,
                    'value': self.DEFAULT_CONTROL_QS,
                    'unit': 'If QS are included in loop'},
                'CorrQS1Current-Mon': {
                    'type': 'float', 'value': 0,
                    'unit': 'A', 'prec': self.DEFAULT_CORR_PREC},
                'CorrQS2Current-Mon': {
                    'type': 'float', 'value': 0,
                    'unit': 'A', 'prec': self.DEFAULT_CORR_PREC},
            })
        dbase = _csdev.add_pvslist_cte(dbase)
        return dbase



"""Const module."""

import time as _time

from mathphys.functions import get_namedtuple as _get_namedtuple

from .. import util as _util
from .. import csdev as _csdev
from ..envars import VACA_PREFIX as _vaca_prefix
from ..devices import DVF as _DVF


class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    NO_YES = ('No', 'Yes')
    STS_LBLS_DVF = ('Not Connected', 'Acquisition Off')


_et = ETypes  # syntactic sugar


class Constants(_csdev.Const):
    """."""

    NoYes = _get_namedtuple('NoYes', _et.NO_YES)
    StsLblsDVF = _get_namedtuple('StsLblsDVF', _et.STS_LBLS_DVF)

    DEF_DVFSTATUS = 0b11111111
    DEF_BEAM_THRESHOLD = 100

    def __init__(self, devname):
        """."""
        self._devname = devname
        self._ioc_prefix = _vaca_prefix + ('-' if _vaca_prefix else '')
        self._ioc_prefix += devname + ':'
        self._database = self._create_database()

    @property
    def devname(self):
        """."""
        return self._devname

    @property
    def ioc_prefix(self):
        """."""
        return self._ioc_prefix

    def get_database(self):
        """."""
        return self._database

    def get_prefix(self):
        """."""
        return self.ioc_prefix

    def _create_database(self):
        database = dict()
        database.update(self._get_image_db())
        database.update(self._get_roi_db())
        database.update(self._get_fit_db())
        database.update(self._get_others_db())
        database = _csdev.add_pvslist_cte(database)

        # rename Properties-Cte to ImgProperties-Cte
        database['ImgProperties-Cte'] = database['Properties-Cte']
        del database['Properties-Cte']

        # TODO: When the known bug related to the use of get_last_commit_hash
        # for determining application version is solved, maybe the value of
        # this PV should be composed in the driver layer, using the values of
        # siriuspy + machine-applicaions versions.
        database['ImgVersion-Cte'] = {
            'type': 'string',
            'value': _util.get_last_commit_hash()
        }

        return database

    def _get_image_db(self):
        dvf_params = _DVF.conv_devname2parameters(self.devname)
        sufix = '-Mon'
        dbase = {
            'ImgDVFSizeX-Cte': {
                'type': 'int', 'unit': 'px'
            },
            'ImgDVFSizeY-Cte': {
                'type': 'int', 'unit': 'px'
            },
            'ImgSizeX' + sufix: {
                'type': 'int', 'unit': 'px'
            },
            'ImgSizeY' + sufix: {
                'type': 'int', 'unit': 'px'
            },
            'ImgProjX' + sufix: {
                'type': 'int', 'count': dvf_params.IMAGE_SIZE_X, 'unit': 'px'
            },
            'ImgProjY' + sufix: {
                'type': 'int', 'count': dvf_params.IMAGE_SIZE_Y, 'unit': 'px'
            },
            'ImgIntensityMin' + sufix: {
                'type': 'int', 'unit': 'intensity'
            },
            'ImgIntensityMax' + sufix: {
                'type': 'int', 'unit': 'intensity'
            },
            'ImgIntensitySum' + sufix: {
                'type': 'int', 'unit': 'intensity'
            },
            'ImgIsSaturated' + sufix: {
                'type': 'enum', 'enums': _et.NO_YES,
                'value': self.NoYes.No,
            },
            'ImgIsWithBeam' + sufix: {
                'type': 'enum', 'enums': _et.NO_YES,
                'value': self.NoYes.No,
            },
            'ImgIsWithBeamThreshold-SP': {
                'type': 'int', 'unit': 'intensity',
                'value': self.DEF_BEAM_THRESHOLD,
            },
            'ImgIsWithBeamThreshold-RB': {
                'type': 'int', 'unit': 'intensity',
                'value': self.DEF_BEAM_THRESHOLD,
            },
        }
        return dbase

    def _get_roi_db(self):
        dvf_params = _DVF.conv_devname2parameters(self.devname)
        db = {}
        rb_ = '-RB'
        sp_ = '-SP'
        mon_ = '-Mon'
        sizes = {'X': dvf_params.IMAGE_SIZE_X, 'Y': dvf_params.IMAGE_SIZE_Y}
        for axis in ['X', 'Y']:
            db.update({
                'ImgROI' + axis + sp_: {
                    'type': 'int', 'count': 2, 'unit': 'px',
                },
                'ImgROI' + axis + rb_: {
                    'type': 'int', 'count': 2, 'unit': 'px'
                },
                'ImgROI' + axis + 'Center' + mon_: {
                    'type': 'int', 'unit': 'px'
                },
                'ImgROI' + axis + 'FWHM' + mon_: {
                    'type': 'int', 'unit': 'px'
                },
                'ImgROI' + axis + 'Proj' + mon_: {
                    'type': 'int', 'count': sizes[axis], 'unit': 'intensity',
                },
                'ImgROI' + axis + 'UpdateWithFWHMFactor-SP': {
                    'type': 'float', 'value': 2.0, 'unit': 'fwhm_factor',
                    'prec': 3, 'lolim': 0, 'hilim': 2000,
                },
                'ImgROI' + axis + 'UpdateWithFWHMFactor-RB': {
                    'type': 'float', 'value': 2.0, 'unit': 'fwhm_factor',
                    'prec': 3, 'lolim': 0, 'hilim': 2000,
                },
            })
        return db

    def _get_fit_db(self):
        db = {}
        mon_ = '-Mon'
        for axis in ['X', 'Y']:
            db.update({
                'ImgROI' + axis + 'FitSigma' + mon_: {
                    'type': 'float', 'unit': 'px', 'prec': 3,
                },
                'ImgROI' + axis + 'FitMean' + mon_: {
                    'type': 'float', 'unit': 'px', 'prec': 3,
                },
                'ImgROI' + axis + 'FitAmplitude' + mon_: {
                    'type': 'float', 'unit': 'intensity', 'prec': 3,
                },
                'ImgROI' + axis + 'FitError' + mon_: {
                    'type': 'float', 'unit': '%', 'prec': 3,
                },
            })
        db.update({
            'ImgROIUpdateWithFWHM-Sel': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl,
            },
            'ImgROIUpdateWithFWHM-Sts': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl,
            },
            'ImgFitAngle' + mon_: {
                'type': 'float', 'unit': 'deg', 'prec': 3,
            },
            'ImgFitSigma1' + mon_: {
                'type': 'float', 'unit': 'px', 'prec': 3,
            },
            'ImgFitSigma2' + mon_: {
                'type': 'float', 'unit': 'px', 'prec': 3,
            },
            'ImgFitProcTime' + mon_: {
                'type': 'float', 'unit': 'ms', 'prec': 3,
            },
            'ImgFitAngleUseCMomSVD-Sel': {
                'type': 'enum', 'enums': _et.NO_YES,
                'value': self.NoYes.Yes,
            },
            'ImgFitAngleUseCMomSVD-Sts': {
                'type': 'enum', 'enums': _et.NO_YES,
                'value': self.NoYes.Yes,
            },
        })
        return db

    def _get_others_db(self):
        db = {}
        db.update({
            'ImgVersion-Cte': {
                'type': 'string',
                'value': _util.get_last_commit_hash()
            },
            'ImgLog-Mon': {
                'type': 'string', 'value': 'Starting...',
            },
            'ImgTimestampBoot-Cte': {
                'type': 'float', 'value': _time.time(),
                'prec': 7, 'unit': 'timestamp',
            },
            'ImgTimestampUpdate-Mon': {
                'type': 'float',
                'prec': 7, 'unit': 'timestamp',
            },
            'ImgDVFReset-Cmd': {
                'type': 'int', 'value': 0,
            },
            'ImgDVFAcquire-Cmd': {
                'type': 'int', 'value': 0,
            },
            'ImgDVFStatus-Mon': {
                'type': 'int', 'value': self.DEF_DVFSTATUS,
            },
            'ImgDVFStatusLabels-Cte': {
                'type': 'string', 'count': len(self.StsLblsDVF._fields),
                'value': self.StsLblsDVF._fields,
            },
        })

        return db



"""TS orbit configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# TS Orbit:
#   -First bpm in the list is the first seen by the beam during injection
#   -Units: um (micrometer)
#
_template_dict = {
    'x': 5*[0.0],  # unit: um
    'y': 5*[0.0]   # unit: um
    }
